"""Functions for exclusive $B\to V\ell^+\ell^-$ decays."""

from math import sqrt, log
from flavio.physics.bdecays.common import meson_quark
from flavio.physics.common import conjugate_par, conjugate_wc
from flavio.physics.bdecays.wilsoncoefficients import wctot_dict
from flavio.config import config
from flavio.physics.running import running
from .amplitudes import *
from flavio.classes import Observable, Prediction
import flavio
import warnings


def dGdq2(J):
    return 3/4. * (2 * J['1s'] + J['1c']) - 1/4. * (2 * J['2s'] + J['2c'])

def dGdq2_ave(J, J_bar):
    return ( dGdq2(J) + dGdq2(J_bar) )/2.

def dGdq2_diff(J, J_bar):
    return ( dGdq2(J) - dGdq2(J_bar) )/2.

# denominator of S_i and A_i observables
def SA_den(J, J_bar):
    return 2*dGdq2_ave(J, J_bar)

# denominator of P_i observables
def P_den(J, J_bar):
    return S_theory_num(J, J_bar, '2s')

def S_theory(J, J_bar, i):
    r"""CP-averaged angular observable $S_i$ in the theory convention."""
    return S_theory_num(J, J_bar, i)/SA_den(J, J_bar)

# numerator
def S_theory_num(J, J_bar, i):
    return (J[i] + J_bar[i])

def A_theory(J, J_bar, i):
    r"""Angular CP asymmetry $A_i$ in the theory convention."""
    return A_theory_num(J, J_bar, i)/SA_den(J, J_bar)

# numerator
def A_theory_num(J, J_bar, i):
    return (J[i] - J_bar[i])

def S_experiment(J, J_bar, i):
    r"""CP-averaged angular observable $S_i$ in the LHCb convention.

    See eq. (C.8) of arXiv:1506.03970v2.
    """
    return S_experiment_num(J, J_bar, i)/SA_den(J, J_bar)

# numerator
def S_experiment_num(J, J_bar, i):
    if i in [4, '6s', '6c', 7, 9]:
        return -S_theory_num(J, J_bar, i)
    return S_theory_num(J, J_bar, i)

def A_experiment(J, J_bar, i):
    r"""Angular CP asymmetry $A_i$ in the LHCb convention.

    See eq. (C.8) of arXiv:1506.03970v2.
    """
    return A_experiment_num(J, J_bar, i)/SA_den(J, J_bar)

# numerator
def A_experiment_num(J, J_bar, i):
    if i in [4, '6s', '6c', 7, 9]:
        return -A_theory_num(J, J_bar, i)
    return A_theory_num(J, J_bar, i)

def AFB_experiment(J, J_bar):
    r"""Forward-backward asymmetry in the LHCb convention.

    See eq. (C.9) of arXiv:1506.03970v2.
    """
    return AFB_experiment_num(J, J_bar)/SA_den(J, J_bar)

def AFB_experiment_num(J, J_bar):
    return 3/4.*S_experiment_num(J, J_bar, '6s')

def AFB_theory(J, J_bar):
    """Forward-backward asymmetry in the original theory convention.
    """
    return AFB_theory_num(J, J_bar)/SA_den(J, J_bar)

def AFB_theory_num(J, J_bar):
    return 3/4.*S_theory_num(J, J_bar, '6s')

def FL(J, J_bar):
    r"""Longitudinal polarization fraction $F_L$"""
    return FL_num(J, J_bar)/SA_den(J, J_bar)

def FL_num(J, J_bar):
    return -S_theory_num(J, J_bar, '2c')

def FLhat(J, J_bar):
    r"""Modified longitudinal polarization fraction for vanishing lepton masses,
    $\hat F_L$.

    See eq. (32) of arXiv:1510.04239.
    """
    return FLhat_num(J, J_bar)/SA_den(J, J_bar)

def FLhat_num(J, J_bar):
    return -S_theory_num(J, J_bar, '1c')


class BVllObservable(object):
    r"""Base class for $B\to V\ell^+\ell^- observable functions that
    facilitates caching/memoization."""

    def __init__(self, B, V, lep, wc_obj, par):
        """Initialize the class and cache results needed more often."""
        self.B = B
        self.V = V
        self.lep = lep
        self.wc_obj = wc_obj
        self.par = par
        self.par_conjugate = conjugate_par(par)
        self.prefactor = prefactor(None, self.par, B, V)
        self.prefactor_conjugate = prefactor(None, self.par_conjugate, B, V)
        self.scale = config['renormalization scale']['bvll']
        self.label = meson_quark[(B,V)] + lep + lep # e.g. bsmumu, bdtautau
        self.wctot_dict = wctot_dict(wc_obj, self.label, self.scale, par)
        self._ff = {}
        self._wceff = {}
        self._wceff_bar = {}
        self._ha = {}
        self._ha_bar = {}
        self._j = {}
        self._j_bar = {}
        self.ml = par['m_'+lep]
        self.mB = par['m_'+B]
        self.mV = par['m_'+V]
        self.mb = running.get_mb(par, self.scale)

    def ff(self, q2):
        """Get form factors. Cache and only recompute if necessary."""
        if q2 not in self._ff:
            self._ff[q2] = get_ff(q2, self.par, self.B, self.V)
        return self._ff[q2]

    def wceff(self, q2):
        """Get effective WCs. Cache and only recompute if necessary."""
        if q2 not in self._wceff:
            self._wceff[q2] = get_wceff(q2, self.wctot_dict, self.par, self.B, self.V, self.lep, self.scale)
        return self._wceff[q2]

    def wceff_bar(self, q2):
        """Get CP conjugate effective WCs. Cache and only recompute if necessary."""
        if q2 not in self._wceff_bar:
            self._wceff_bar[q2] = get_wceff(q2, conjugate_wc(self.wctot_dict), self.par_conjugate, self.B, self.V, self.lep, self.scale)
        return self._wceff_bar[q2]

    def helicity_amps_ff(self, q2, cp_conjugate):
        """Get helicity amps proportional to FFs. Cache and only recompute if necessary."""
        if not cp_conjugate:
            return angular.helicity_amps_v(q2,
                     self.mB, self.mV, self.mb, 0, self.ml, self.ml,
                     self.ff(q2), self.wceff(q2), self.prefactor)
        else:
            return angular.helicity_amps_v(q2,
                     self.mB, self.mV, self.mb, 0, self.ml, self.ml,
                     self.ff(q2), self.wceff_bar(q2), self.prefactor_conjugate)

    def ha(self, q2):
        """Get full helicity amps. Cache and only recompute if necessary."""
        if q2 not in self._ha:
            self._ha[q2] = add_dict((
                self.helicity_amps_ff(q2, cp_conjugate=False),
                get_ss(q2, self.wc_obj, self.par, self.B, self.V, cp_conjugate=False),
                get_subleading(q2, self.wc_obj, self.par, self.B, self.V, cp_conjugate=False)
                ))
        return self._ha[q2]

    def ha_bar(self, q2):
        """Get CP conjugate full helicity amps. Cache and only recompute if necessary."""
        if q2 not in self._ha_bar:
            self._ha_bar[q2] = add_dict((
                self.helicity_amps_ff(q2, cp_conjugate=True),
                get_ss(q2, self.wc_obj, self.par, self.B, self.V, cp_conjugate=True),
                get_subleading(q2, self.wc_obj, self.par, self.B, self.V, cp_conjugate=True)
                ))
        return self._ha_bar[q2]

    def j(self, q2):
        """Get angular coeffs. Cache and only recompute if necessary."""
        h = self.ha(q2)
        if q2 not in self._j:
            self._j[q2] = angular.angularcoeffs_general_v(h, q2, self.mB, self.mV, self.mb, 0, self.ml, self.ml)
        return self._j[q2]

    def jbar(self, q2):
        """Get CP conjugate angular coeffs. Cache and only recompute if necessary."""
        hbar = self.ha_bar(q2)
        if q2 not in self._j_bar:
            self._j_bar[q2] = angular.angularcoeffs_general_v(hbar, q2, self.mB, self.mV, self.mb, 0, self.ml, self.ml)
        return self._j_bar[q2]

    def jfunc(self, function, q2):
        """Return a function of J and Jbar at one value of q2."""
        return function(self.j(q2), self.jbar(q2))


class BVllObservableDifferential(BVllObservable):
    """Base class for differential observables depending on q2."""
    def __init__(self, q2, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.q2 = q2
        if q2 < 4 * self.ml**2 or q2 > (self.mB - self.mV)**2:
            self.allowed = False
        else:
            self.allowed = True


class BVllObservableBinned(BVllObservable):
    """Base class for binned observables depending on q2min and q2max."""
    def __init__(self, q2min, q2max, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.q2min = q2min
        self.q2max = q2max
        self.q2min_allowed = max(4 * self.ml**2, self.q2min)
        self.q2max_allowed = min((self.mB - self.mV)**2, self.q2max)


class BVll_dBRdq2(BVllObservableDifferential):
    """Differential branching ratio"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self):
        if not self.allowed:
            return 0
        tauB = self.par['tau_' + self.B]
        return tauB * self.jfunc(dGdq2_ave, self.q2)


class BVll_obs(BVllObservableDifferential):
    """Differential function of angular coefficients"""
    def __init__(self, func, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.func = func

    def __call__(self):
        if not self.allowed:
            return 0
        return self.jfunc(self.func, self.q2)


class BVll_ratio(BVllObservableDifferential):
    """Differential ratio of functions of angular coefficients"""
    def __init__(self, func_num, func_den, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.func_num = func_num
        self.func_den = func_den

    def __call__(self):
        if not self.allowed:
            return 0
        num = self.jfunc(self.func_num, self.q2)
        if num == 0:
            return 0
        den = self.jfunc(self.func_den, self.q2)
        return num / den


class BVll_pprime(BVllObservableDifferential):
    r"""Differential $P'$ observables"""
    def __init__(self, func_num, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.func_num = func_num

    @staticmethod
    def func_2s(J, J_bar):
        return S_experiment_num(J, J_bar, '2s')

    @staticmethod
    def func_2c(J, J_bar):
        return S_experiment_num(J, J_bar, '2c')

    def __call__(self):
        if not self.allowed:
            return 0
        num = self.jfunc(self.func_num, self.q2)
        if num == 0:
            return 0
        den_2s = self.jfunc(self.func_2s, self.q2)
        den_2c = self.jfunc(self.func_2c, self.q2)
        den = 2 * sqrt(-den_2s * den_2c)
        return num / den


class BVll_dBRdq2_int(BVllObservableBinned):
    """Binned branching ratio"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.epsrel = 0.005

    def obs(self, q2):
        tauB = self.par['tau_' + self.B]
        return tauB * self.jfunc(dGdq2_ave, q2)

    def __call__(self):
        if self.q2max_allowed <= self.q2min_allowed:
            return 0
        return nintegrate_pole(self.obs, self.q2min_allowed, self.q2max_allowed, epsrel=self.epsrel) / (self.q2max - self.q2min)


class BVll_obs_int(BVllObservableBinned):
    """Binned function of angular coefficients"""
    def __init__(self, func, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.epsrel = 0.005
        self.func = func

    def obs(self, q2):
        return self.jfunc(self.func, q2)

    def __call__(self):
        if self.q2max_allowed <= self.q2min_allowed:
            return 0
        return nintegrate_pole(self.obs, self.q2min_allowed, self.q2max_allowed, epsrel=self.epsrel) / (self.q2max - self.q2min)


class BVll_int_ratio(BVllObservableBinned):
    """Binned ratio of functions if angular coefficients"""
    def __init__(self, func_num, func_den, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.epsrel = 0.005
        self.func_num = func_num
        self.func_den = func_den

    def obs_num(self, q2):
        return self.jfunc(self.func_num, q2)

    def obs_den(self, q2):
        return self.jfunc(self.func_den, q2)

    def __call__(self):
        if self.q2max_allowed <= self.q2min_allowed:
            return 0
        num = nintegrate_pole(self.obs_num, self.q2min_allowed, self.q2max_allowed, epsrel=self.epsrel)
        if num == 0:
            return 0
        den = nintegrate_pole(self.obs_den, self.q2min_allowed, self.q2max_allowed, epsrel=self.epsrel)
        return num / den


class BVll_int_pprime(BVllObservableBinned):
    r"""Binned $P'$ observables"""
    def __init__(self, func_num, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.epsrel = 0.005
        self.func_num = func_num

    def obs_num(self, q2):
        return self.jfunc(self.func_num, q2)

    def obs_2s(self, q2):
        return self.jfunc(lambda J, J_bar: S_experiment_num(J, J_bar, '2s'), q2)

    def obs_2c(self, q2):
        return self.jfunc(lambda J, J_bar: S_experiment_num(J, J_bar, '2c'), q2)

    def __call__(self):
        if self.q2max_allowed <= self.q2min_allowed:
            return 0
        num = nintegrate_pole(self.obs_num, self.q2min_allowed, self.q2max_allowed, epsrel=self.epsrel)
        if num == 0:
            return 0
        den_2s = nintegrate_pole(self.obs_2s, self.q2min_allowed, self.q2max_allowed, epsrel=self.epsrel)
        den_2c = nintegrate_pole(self.obs_2c, self.q2min_allowed, self.q2max_allowed, epsrel=self.epsrel)
        den = 2 * sqrt(-den_2s * den_2c)
        return num / den


def nintegrate_pole(function, q2min, q2max, epsrel=0.005):
    # this is a special integration function to treat the presence of the
    # photon pole at low q^2. If q2min is below 0.1 GeV^2, it adds and subtracts
    # the 1/q^2-enhanced pole part to split the integral into a well-behaved part
    # and one that is trivially solved analytically.
    # This leads to a huge speed-up.
    if q2min <= 0.1 and q2min > 0:
        q20 = q2min
        f_q20 = function(q20)
        int_a = flavio.math.integrate.nintegrate(lambda q2: function(q2)-f_q20*q20/q2, q2min, q2max, epsrel=epsrel)
        int_b = f_q20*q20 * log(q2max/q2min)
        return int_a + int_b
    else:
        return flavio.math.integrate.nintegrate(function, q2min, q2max)



# Functions returning functions needed for Prediction instances

def bvll_obs_int_ratio_leptonflavour(func, B, V, l1, l2):
    def fct(wc_obj, par, q2min, q2max):
        # ignore QCDF warnings for LFU ratios!
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="The QCDF corrections should not be trusted .*")
            numobj = BVll_obs_int(func, q2min, q2max, B, V, l1, wc_obj, par)
            numobj.epsrel = 0.0005
            num = numobj()
            if num == 0:
                return 0
            denobj = BVll_obs_int(func, q2min, q2max, B, V, l2, wc_obj, par)
            denobj.epsrel = 0.0005
            den = denobj()
            return num / den
    return fct

def bvll_obs_ratio_leptonflavour(func, B, V, l1, l2):
    def fct(wc_obj, par, q2):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="The QCDF corrections should not be trusted .*")
            num = BVll_obs(func, q2, B, V, l1, wc_obj, par)()
            if num == 0:
                return 0
            den = BVll_obs(func, q2, B, V, l2, wc_obj, par)()
            return num / den
    return fct


# Observable and Prediction instances

_tex = {'e': 'e', 'mu': '\mu', 'tau': r'\tau'}
_observables = {
'ACP': {'func_num': dGdq2_diff, 'tex': r'A_\text{CP}', 'desc': 'Direct CP asymmetry'},
'AFB': {'func_num': AFB_experiment_num, 'tex': r'A_\text{FB}', 'desc': 'forward-backward asymmetry'},
'FL': {'func_num': FL_num, 'tex': r'F_L', 'desc': 'longitudinal polarization fraction'},
'S3': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 3), 'tex': r'S_3', 'desc': 'CP-averaged angular observable'},
'S4': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 4), 'tex': r'S_4', 'desc': 'CP-averaged angular observable'},
'S5': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 5), 'tex': r'S_5', 'desc': 'CP-averaged angular observable'},
'S6c': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, '6c'), 'tex': r'S_6^c', 'desc': 'CP-averaged angular observable'},
'S7': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 7), 'tex': r'S_7', 'desc': 'CP-averaged angular observable'},
'S8': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 8), 'tex': r'S_8', 'desc': 'CP-averaged angular observable'},
'S9': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 9), 'tex': r'S_9', 'desc': 'CP-averaged angular observable'},
'A3': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 3), 'tex': r'A_3', 'desc': 'Angular CP asymmetry'},
'A4': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 4), 'tex': r'A_4', 'desc': 'Angular CP asymmetry'},
'A5': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 5), 'tex': r'A_5', 'desc': 'Angular CP asymmetry'},
'A6s': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, '6s'), 'tex': r'A_6^s', 'desc': 'Angular CP asymmetry'},
'A7': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 7), 'tex': r'A_7', 'desc': 'Angular CP asymmetry'},
'A8': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 8), 'tex': r'A_8', 'desc': 'Angular CP asymmetry'},
'A9': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 9), 'tex': r'A_9', 'desc': 'Angular CP asymmetry'},
}
# for the P observables, the convention of LHCb is used. This differs by a
# sign in P_2 and P_3 from the convention in arXiv:1303.5794
_observables_p = {
'P1': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 3)/2., 'tex': r'P_1', 'desc': "CP-averaged \"optimized\" angular observable"},
'P2': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, '6s')/8., 'tex': r'P_2', 'desc': "CP-averaged \"optimized\" angular observable"},
'P3': {'func_num': lambda J, J_bar: -S_experiment_num(J, J_bar, 9)/4., 'tex': r'P_3', 'desc': "CP-averaged \"optimized\" angular observable"},
'ATIm': {'func_num': lambda J, J_bar: A_experiment_num(J, J_bar, 9)/2., 'tex': r'A_T^\text{Im}', 'desc': "Transverse CP asymmetry"},
}
_observables_pprime = {
'P4p': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 4), 'tex': r'P_4^\prime', 'desc': "CP-averaged \"optimized\" angular observable"},
'P5p': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 5), 'tex': r'P_5^\prime', 'desc': "CP-averaged \"optimized\" angular observable"},
# yes, P6p depends on J_7, not J_6. Don't ask why.
'P6p': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 7), 'tex': r'P_6^\prime', 'desc': "CP-averaged \"optimized\" angular observable"},
'P8p': {'func_num': lambda J, J_bar: S_experiment_num(J, J_bar, 8), 'tex': r'P_8^\prime', 'desc': "CP-averaged \"optimized\" angular observable"},
}
_hadr = {
'B0->K*': {'tex': r"B^0\to K^{\ast 0}", 'B': 'B0', 'V': 'K*0', },
'B+->K*': {'tex': r"B^+\to K^{\ast +}", 'B': 'B+', 'V': 'K*+', },
}

def make_metadata_binned(M, l, obs, obsdict):
    _process_tex = _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+r"^-"
    _process_taxonomy = r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _process_tex + r"$"
    B = _hadr[M]['B']
    V = _hadr[M]['V']
    _obs_name = "<" + obs + ">("+M+l+l+")"
    _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
    _obs.set_description('Binned ' + obsdict['desc'] + r" in $" + _process_tex + r"$")
    _obs.tex = r"$\langle " + obsdict['tex'] + r"\rangle(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    return _obs

def make_metadata_differential(M, l, obs, obsdict):
    _process_tex = _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+r"^-"
    _process_taxonomy = r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _process_tex + r"$"
    B = _hadr[M]['B']
    V = _hadr[M]['V']
    _obs_name = obs + "("+M+l+l+")"
    _obs = Observable(name=_obs_name, arguments=['q2'])
    _obs.set_description(obsdict['desc'][0].capitalize() + obsdict['desc'][1:] + r" in $" + _process_tex + r"$")
    _obs.tex = r"$" + obsdict['tex'] + r"(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    return _obs

def make_obs(M, l, obs, obsdict):
    B = _hadr[M]['B']
    V = _hadr[M]['V']
    func_num = obsdict['func_num']

    # binned angular observables
    _obs = make_metadata_binned(M, l, obs, obsdict)
    func = lambda wc_obj, par, q2min, q2max: BVll_int_ratio(func_num, SA_den, q2min, q2max, B, V, l, wc_obj, par)()
    Prediction(_obs.name, func)

    # differential angular observables
    _obs = make_metadata_differential(M, l, obs, obsdict)
    func = lambda wc_obj, par, q2: BVll_ratio(func_num, SA_den, q2, B, V, l, wc_obj, par)()
    Prediction(_obs.name, func)


def make_obs_p(M, l, obs, obsdict):
    B = _hadr[M]['B']
    V = _hadr[M]['V']
    func_num = obsdict['func_num']

    # binned "optimized" angular observables P
    _obs = make_metadata_binned(M, l, obs, obsdict)
    func = lambda wc_obj, par, q2min, q2max: BVll_int_ratio(func_num, P_den, q2min, q2max, B, V, l, wc_obj, par)()
    Prediction(_obs.name, func)

    # differential "optimized"  angular observables
    _obs = make_metadata_differential(M, l, obs, obsdict)
    func = lambda wc_obj, par, q2: BVll_ratio(func_num, P_den, q2, B, V, l, wc_obj, par)()
    Prediction(_obs.name, func)


def make_obs_pprime(M, l, obs, obsdict):
    B = _hadr[M]['B']
    V = _hadr[M]['V']
    func_num = obsdict['func_num']

    # binned "optimized"  angular observables
    _obs = make_metadata_binned(M, l, obs, obsdict)
    func = lambda wc_obj, par, q2min, q2max: BVll_int_pprime(func_num, q2min, q2max, B, V, l, wc_obj, par)()
    Prediction(_obs.name, func)

    # differential "optimized"  angular observables
    _obs = make_metadata_differential(M, l, obs, obsdict)
    func = lambda wc_obj, par, q2: BVll_pprime(func_num, q2, B, V, l, wc_obj, par)()
    Prediction(_obs.name, func)


def make_obs_br(M, l):
    """Make observable instances for branching ratios"""
    _process_tex = _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+r"^-"
    _process_taxonomy = r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _process_tex + r"$"
    B = _hadr[M]['B']
    V = _hadr[M]['V']

    # binned branching ratio
    _obs_name = "<dBR/dq2>("+M+l+l+")"
    _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
    _obs.set_description(r"Binned differential branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\langle \frac{d\text{BR}}{dq^2} \rangle(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    func = lambda wc_obj, par, q2min, q2max: BVll_dBRdq2_int(q2min, q2max, B, V, l, wc_obj, par)()
    Prediction(_obs_name, func)

    # differential branching ratio
    _obs_name = "dBR/dq2("+M+l+l+")"
    _obs = Observable(name=_obs_name, arguments=['q2'])
    _obs.set_description(r"Differential branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\frac{d\text{BR}}{dq^2}(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    func = lambda wc_obj, par, q2: BVll_dBRdq2(q2, B, V, l, wc_obj, par)()
    Prediction(_obs_name, func)


def make_obs_lfur(M, l):
    """Make observable instances for lepton flavour ratios"""
    # binned ratio of BRs
    _obs_name = "<R"+l[0]+l[1]+">("+M+"ll)"
    _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
    _obs.set_description(r"Ratio of partial branching ratios of $" + _hadr[M]['tex'] +_tex[l[0]]+r"^+ "+_tex[l[0]]+r"^-$" + " and " + r"$" + _hadr[M]['tex'] +_tex[l[1]]+r"^+ "+_tex[l[1]]+"^-$")
    _obs.tex = r"$\langle R_{" + _tex[l[0]] + ' ' + _tex[l[1]] + r"} \rangle(" + _hadr[M]['tex'] + r"\ell^+\ell^-)$"
    for li in l:
        # add taxonomy for both processes (e.g. B->Vee and B->Vmumu)
        _obs.add_taxonomy(r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _hadr[M]['tex'] +_tex[li]+r"^+"+_tex[li]+r"^-$")
    Prediction(_obs_name, bvll_obs_int_ratio_leptonflavour(dGdq2_ave, _hadr[M]['B'], _hadr[M]['V'], *l))

    # differential ratio of BRs
    _obs_name = "R"+l[0]+l[1]+"("+M+"ll)"
    _obs = Observable(name=_obs_name, arguments=['q2'])
    _obs.set_description(r"Ratio of differential branching ratios of $" + _hadr[M]['tex'] +_tex[l[0]]+r"^+ "+_tex[l[0]]+r"^-$" + " and " + r"$" + _hadr[M]['tex'] +_tex[l[1]]+r"^+ "+_tex[l[1]]+"^-$")
    _obs.tex = r"$R_{" + _tex[l[0]] + ' ' + _tex[l[1]] + r"} (" + _hadr[M]['tex'] + r"\ell^+\ell^-)$"
    for li in l:
        # add taxonomy for both processes (e.g. B->Vee and B->Vmumu)
        _obs.add_taxonomy(r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _hadr[M]['tex'] +_tex[li]+r"^+"+_tex[li]+r"^-$")
    func = lambda wc_obj, par, q2: BVll_dBRdq2(q2, B, V, l, wc_obj, par)()
    Prediction(_obs_name, bvll_obs_ratio_leptonflavour(dGdq2_ave, _hadr[M]['B'], _hadr[M]['V'], *l))

# loop over all cases
for l in ['e', 'mu', 'tau']:
    for M in _hadr.keys():
        for obs, obsdict in _observables.items():
            # angular obs
            make_obs(M, l, obs, obsdict)
        for obs, obsdict in _observables_p.items():
            # P obs
            make_obs_p(M, l, obs, obsdict)
        for obs, obsdict in _observables_pprime.items():
            # P' obs
            make_obs_pprime(M, l, obs, obsdict)
        # BRs
        make_obs_br(M, l)

# lepton flavour ratios
for l in [('mu','e'), ('tau','mu'),]:
    for M in _hadr.keys():
        make_obs_lfur(M, l)



"""Default configuration options for `matplotlib`"""

from matplotlib import rc, rcParams

rc('text', usetex=True)
rc('font',**{'family':'serif',
             'serif':['Computer Modern Roman','cmu serif']+rcParams['font.serif']})
rcParams['xtick.labelsize'] = 10
rcParams['ytick.labelsize'] = 10
rcParams['axes.labelsize'] = 14
rcParams['text.latex.preamble']=[r"\usepackage{amsmath}"]



from math import sqrt
import numpy as np
from flavio.physics.bdecays.formfactors.common import z

def zs(mX, mP, par, q2):
    mLb = par['m_Lambdab']
    mB = par['m_B+']
    zq2 = z(mB, mP, q2, t0=(mLb - mX)**2) # cf. eq. (34) of arXiv:1602.01399
    return np.array([1, zq2, zq2**2])

def pole(ff, mres, q2):
    m = mres[_mresdict[ff]]
    return 1/(1-q2/m**2)

_mresdict = {'fV0': '1-',
             'fVperp': '1-',
             'fT0': '1-',
             'fTperp': '1-',
             'fVt': '0+',
             'fA0': '1+',
             'fAperp': '1+',
             'fT50': '1+',
             'fT5perp': '1+',
             'fAt': '0-',
            }
# resonance masses used in arXiv:1602.01399
_mres = {}
_mres['b->s'] = {'1-': 5.416, '0+': 5.711, '1+': 5.750, '0-': 5.367}

# X: name of the final-state baryon, q: quark-level transition (e.g. 'b->s'),
# P: name of the pseudoscalar meson with the same flavour QN as X (needed for z)
_process_dict = {}
_process_dict['Lambdab->Lambda'] =    {'X': 'Lambda', 'P': 'K+', 'q': 'b->s'}

def ff(process, q2, par, n=2):
    r"""Central value of $\Lambda_b\to X_{1/2}$ form factors in the helicity
    basis and simplified series expansion (SSE) parametrization.

    The helicity basis defines the form factors
    $f_+$, $f_0$, $f_\perp$, $g_+$, $g_0$, $g_\perp$,
    $h_+$, $h_\perp$, $\tilde h_+$, $\tilde h_\perp$,

    The SSE defines
    $$F_i(q^2) = P_i(q^2) \sum_k a_k^i \,z(q^2)^k$$
    where $P_i(q^2)=(1-q^2/m_{R,i}^2)^{-1}$ is a simple pole.
    """
    pd = _process_dict[process]
    mres = _mres[pd['q']]
    mX = par['m_'+pd['X']]
    mP = par['m_'+pd['P']]
    ff = {}
    # implementing the two endpoint relations in (7) and (8) of arXiv:1602.01399
    par[process+' SSE a0_fAperp'] = par[process+' SSE a0_fA0']
    par[process+' SSE a0_fT5perp'] = par[process+' SSE a0_fT50']
    for i in ['fA0', 'fAperp', 'fAt', 'fT0', 'fT50', 'fT5perp', 'fTperp', 'fV0', 'fVperp', 'fVt']:
        a = [ par[process+' SSE ' + 'a' + str(j) + '_' + i] for j in range(n) ]
        ff[i] = pole(i, mres, q2)*np.dot(a, zs(mX, mP, par, q2)[:n])
    return ff



r"""Functions for $K\to\ell\nu$ and $\pi\to\ell\nu$ decays."""

import flavio
from flavio.physics.bdecays.blnu import br_plnu_general
from math import pi, log
from flavio.math.functions import li2
from flavio.physics.taudecays.taulnunu import GFeff


def br_plnu(wc_obj, par, P, lep):
    r"""Branching ratio of $P^+\to\ell^+\nu_\ell$."""
    return sum([ _br_plnu(wc_obj, par, P, lep, nu) for nu in ['e','mu','tau']])

def _br_plnu(wc_obj, par, P, lep, nu):
    # CKM element
    if P == 'K+':
        Vij = flavio.physics.ckm.get_ckm(par)[0, 1]
        qiqj = 'su'
    elif P == 'pi+':
        Vij = flavio.physics.ckm.get_ckm(par)[0, 0]
        qiqj = 'du'
    # renormalization scale is m_rho
    scale = par['m_rho0']
    # Wilson coefficients
    wc = wc_obj.get_wc(qiqj + lep + 'nu' + nu, scale, par, nf_out=3)
    # add SM contribution to Wilson coefficient
    if lep == nu:
        # for the SM contribution, need the Fermi constant with possible
        # NP effects in mu->enunu subtracted, not the measured one
        r_GF = GFeff(wc_obj, par) / par['GF']
        wc['CVL_'+qiqj+lep+'nu'+nu] += flavio.physics.bdecays.wilsoncoefficients.get_CVLSM(par, scale, nf=3) * r_GF
    if P == 'K+':
        mq = flavio.physics.running.running.get_ms(par, scale)
    elif P == 'pi+':
        mq = flavio.physics.running.running.get_md(par, scale)
    mu = flavio.physics.running.running.get_mu(par, scale)
    return br_plnu_general(wc, par, Vij, P, qiqj, lep, nu, mq, mu, delta=delta_Plnu(par, P, lep))

def r_plnu(wc_obj, par, P):
    # resumming logs according to (111) of 0707.4464
    # (this is negligibly small for the individual rates)
    rg_corr = 1.00055
    return rg_corr*br_plnu(wc_obj, par, P, 'e')/br_plnu(wc_obj, par, P, 'mu')

def gamma_plnu(wc_obj, par, P, l):
    return br_plnu(wc_obj, par, P, l) / par['tau_' + P]


def delta_Plnu(par, P, lep):
    mrho = par['m_rho0']
    mP = par['m_'+P]
    ml = par['m_'+lep]
    alpha_e = 1/137.035999139 # this is alpha_e(0), a constant for our purposes
    c1 = par['c1_'+P+'lnu'] # e.g. c1_K+lnu
    c2 = par['c2_'+P+'lnu']
    c3 = par['c3_'+P+'lnu']
    if lep=='mu':
        c4 = par['c4_'+P+'munu']
    elif lep=='e':
        c4 = 0 # c_4 tends to zero with vanishing lepton mass
    c2t = par['c2t_'+P+'lnu']
    return alpha_e/pi * (F(ml**2/mP**2) - 3/2.*log(mrho/mP) - c1
            - ml**2/mrho**2 * (c2 * log(mrho**2/ml**2) + c3 + c4)
            + mP**2/mrho**2 * c2t * log(mrho**2/ml**2) )

def F(z):
    return ( 3/2.*log(z) + (13-19*z)/(8*(1-z))
            - (8-5*z)/(4*(1-z)**2)*z*log(z)
            - ((1+z)/(1-z)*log(z)+2)*log(1-z)
            -2*(1+z)/(1-z)*li2(1-z) )

# function returning function needed for prediction instance
def br_plnu_fct(P, lep):
    def f(wc_obj, par):
        return br_plnu(wc_obj, par, P, lep)
    return f

def r_plnu_fct_p(P):
    def f(wc_obj, par):
        return r_plnu(wc_obj, par, P)
    return f

def gamma_plnu_fct(P, l):
    def f(wc_obj, par):
        return gamma_plnu(wc_obj, par, P, l)
    return f

def r_plnu_fct_l(l):
    def f(wc_obj, par):
        return gamma_plnu(wc_obj, par, 'K+', l) / gamma_plnu(wc_obj, par, 'pi+', l)
    return f

# Observable and Prediction instances

_tex = {'e': 'e', 'mu': r'\mu'}
_tex_p = {'K+': r'K^+', 'pi+': r'\pi^+',}

# K+->lnu
for l in ['e', 'mu']:
    # Individual (e and mu) modes
    _obs_name = "BR(K+->"+l+"nu)"
    _obs = flavio.classes.Observable(_obs_name)
    _process_tex = r"K^+\to "+_tex[l]+r"^+\nu"
    _process_taxonomy = r'Process :: $s$ hadron decays :: Leptonic tree-level decays :: $K\to \ell\nu$ :: $' + _process_tex + r'$'
    _obs.add_taxonomy(_process_taxonomy)
    _obs.set_description(r"Branching ratio of $" + _process_tex +r"(\gamma)$")
    _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
    flavio.classes.Prediction(_obs_name, br_plnu_fct('K+', l))

    # pi/K ratio
    _obs_name = "RKpi(P+->"+l+"nu)"
    _obs = flavio.classes.Observable(_obs_name)
    _process_tex_1 = r"\pi^+\to "+_tex[l]+r"^+\nu"
    _process_tex_2 = r"K^+\to "+_tex[l]+r"^+\nu"
    _process_taxonomy = r'Process :: Unflavoured meson decays :: Leptonic tree-level decays :: $\pi\to \ell\nu$ :: $' + _process_tex_1 + r'$'
    _obs.add_taxonomy(_process_taxonomy)
    _process_taxonomy = r'Process :: $s$ hadron decays :: Leptonic tree-level decays :: $K\to \ell\nu$ :: $' + _process_tex_2 + r'$'
    _obs.add_taxonomy(_process_taxonomy)
    _obs.set_description(r"Ratio $" + _process_tex_2 +r"$ and $" + _process_tex_1 +r"$ decay rates")
    _obs.tex = r"$\text{R}_{K\pi}(P^+\to "+_tex[l]+r"^+\nu)$"
    flavio.classes.Prediction(_obs_name, r_plnu_fct_l(l))

# e/mu ratios
_obs_name = "Remu(K+->lnu)"
_obs = flavio.classes.Observable(_obs_name)
_obs.set_description(r"Ratio of branching ratios of $K^+\to e^+\nu_e$ and $K^+\to \mu^+\nu_\mu$")
_obs.tex = r"$R_{e\mu}(K^+\to \ell^+\nu)$"
_obs.add_taxonomy(r'Process :: $s$ hadron decays :: Leptonic tree-level decays :: $K\to \ell\nu$ :: $K^+\to e^+\nu$')
_obs.add_taxonomy(r'Process :: $s$ hadron decays :: Leptonic tree-level decays :: $K\to \ell\nu$ :: $K^+\to \mu^+\nu$')
flavio.classes.Prediction(_obs_name, r_plnu_fct_p('K+'))

# for the pion decay, the only branching ratio needed is pi->enu, as
# pi->munu is almost 100%!
_obs_name = "BR(pi+->enu)"
_obs = flavio.classes.Observable(_obs_name)
_process_tex = r"\pi^+\to e^+\nu"
_obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
_obs.tex = r"$\text{BR}(" + _process_tex + r")$"
_process_taxonomy = r'Process :: Unflavoured meson decays :: Leptonic tree-level decays :: $\pi\to \ell\nu$ :: $' + _process_tex + r'$'
_obs.add_taxonomy(_process_taxonomy)
flavio.classes.Prediction(_obs_name, r_plnu_fct_p('pi+'))

# include the partial width instead of BR for pi->munu
_obs_name = "Gamma(pi+->munu)"
_obs = flavio.classes.Observable(_obs_name)
_process_tex = r"\pi^+\to \mu^+\nu"
_obs.set_description(r"Decay rate of $" + _process_tex + r"$")
_obs.tex = r"$\Gamma(" + _process_tex + r")$"
_process_taxonomy = r'Process :: Unflavoured meson decays :: Leptonic tree-level decays :: $\pi\to \ell\nu$ :: $' + _process_tex + r'$'
_obs.add_taxonomy(_process_taxonomy)
flavio.classes.Prediction(_obs_name, gamma_plnu_fct('pi+', 'mu'))



r"""Functions for lepton flavour violating $\tau\to \ell_1\ell_2\ell_3$ decays."""

import flavio
from math import pi, log, sqrt


# names of LFV sectors in WCxf
wcxf_sector_names = {('tau', 'mu'): 'mutau',
                     ('tau', 'e'): 'taue',
                     ('mu', 'e'): 'mue',
                     ('tau', 'e', 'mu', 'e'): 'etauemu',
                     ('tau', 'mu', 'e', 'mu'): 'muemutau', }


def _BR_taumuee(mtau, me, wc):
    # (22) of hep-ph/0404211
    return (abs(wc['CVLL'])**2 + abs(wc['CVLR'])**2 + abs(wc['CVRL'])**2 + abs(wc['CVRR'])**2
            + 1 / 4 * (abs(wc['CSLL'])**2 + abs(wc['CSLR'])**2 + abs(wc['CSRL'])**2 + abs(wc['CSRR'])**2)
            + 12 * (abs(wc['CTLL'])**2 + abs(wc['CTRR'])**2)
            + 8 * (wc['C7'] * (wc['CVLL'] + wc['CVLR']).conjugate()
                 + wc['C7p'] * (wc['CVRL'] + wc['CVRR']).conjugate()).real
            + 32 * (abs(wc['C7'])**2 + abs(wc['C7p'])**2) * (log(mtau**2 / me**2) - 3))


def _BR_tau3mu(mtau, mmu, wc):
    # (23) of hep-ph/0404211
    # (117) of hep-ph/9909265
    return (2 * abs(wc['CVLL'])**2 + abs(wc['CVLR'])**2 + abs(wc['CVRL'])**2 + 2 * abs(wc['CVRR'])**2
            + 1 / 8 * (abs(wc['CSLL'])**2 + abs(wc['CSRR'])**2)
            + 8 * (wc['C7'] * (2 * wc['CVLL'] + wc['CVLR']).conjugate()
                          + wc['C7p'] * (wc['CVRL'] + 2 * wc['CVRR']).conjugate()).real
            + 32 * (abs(wc['C7'])**2 + abs(wc['C7p'])**2)
            * (log(mtau**2 / mmu**2) - 11 / 4))


def _BR_taumuemu(wc):
    r"""Function for $\Delta L=2$ decays like $\tau^-\to \mu^- e^+ \mu^-$."""
    return (2 * abs(wc['CVLL'])**2 + abs(wc['CVLR'])**2 + abs(wc['CVRL'])**2 + 2 * abs(wc['CVRR'])**2
            + 1 / 8 * (abs(wc['CSLL'])**2 + abs(wc['CSRR'])**2))


def wc_eff(wc_obj, par, scale, l0, l1, l2, l3, nf_out=4):
    r"""Get the effective Wilson coefficients for the $l_0^-\to l_1^-l_2^+l_3^-$
    transition as a dictionary."""
    if l2 == l3:
        sector = wcxf_sector_names[l0, l1]
    else:
        sector = wcxf_sector_names[l0, l1, l2, l3]
    alpha = flavio.physics.running.running.get_alpha_e(par, scale, nf_out=4)
    e = sqrt(4 * pi * alpha)
    ml0 = par['m_' + l0]
    wc = wc_obj.get_wc(sector, scale, par, nf_out=nf_out)
    wceff = {}
    if (l0, l1, l2, l3) == ('tau', 'mu', 'mu', 'mu'):
        wceff['C7'] = e / ml0 * wc['Cgamma_taumu']
        wceff['C7p'] = e / ml0 * wc['Cgamma_mutau'].conjugate()
        wceff['CVLL'] = wc['CVLL_mumutaumu']
        wceff['CVLR'] = wc['CVLR_taumumumu']
        wceff['CVRL'] = wc['CVLR_mumutaumu']
        wceff['CVRR'] = wc['CVRR_mumutaumu']
        wceff['CSRR'] = wc['CSRR_mumutaumu']
        wceff['CSLL'] = wc['CSRR_mumumutau'].conjugate()
    elif (l0, l1, l2, l3) == ('tau', 'e', 'e', 'e'):
        wceff['C7'] = e / ml0 * wc['Cgamma_taue']
        wceff['C7p'] = e / ml0 * wc['Cgamma_etau'].conjugate()
        wceff['CVLL'] = wc['CVLL_eetaue']
        wceff['CVLR'] = wc['CVLR_taueee']
        wceff['CVRL'] = wc['CVLR_eetaue']
        wceff['CVRR'] = wc['CVRR_eetaue']
        wceff['CSRR'] = wc['CSRR_eetaue']
        wceff['CSLL'] = wc['CSRR_eeetau'].conjugate()
    elif (l0, l1, l2, l3) == ('tau', 'mu', 'e', 'e'):
        wceff['C7'] = e / ml0 * wc['Cgamma_taumu']
        wceff['C7p'] = e / ml0 * wc['Cgamma_mutau'].conjugate()
        wceff['CVLL'] = wc['CVLL_eetaumu']
        wceff['CVLR'] = wc['CVLR_taumuee']
        wceff['CVRL'] = wc['CVLR_eetaumu']
        wceff['CVRR'] = wc['CVRR_eetaumu']
        wceff['CSRR'] = wc['CSRR_eetaumu'] - wc['CSRR_taueemu'] / 2
        wceff['CSLL'] = wc['CSRR_eemutau'].conjugate() - wc['CSRR_mueetau'].conjugate() / 2
        wceff['CSLR'] = -2 * wc['CVLR_taueemu']
        wceff['CSRL'] = -2 * wc['CVLR_mueetau'].conjugate()
        wceff['CTLL'] = -wc['CSRR_mueetau'].conjugate() / 8
        wceff['CTRR'] = -wc['CSRR_taueemu'] / 8
    elif (l0, l1, l2, l3) == ('tau', 'e', 'mu', 'mu'):
        wceff['C7'] = e / ml0 * wc['Cgamma_taue']
        wceff['C7p'] = e / ml0 * wc['Cgamma_etau'].conjugate()
        wceff['CVLL'] = wc['CVLL_muetaumu']
        wceff['CVLR'] = wc['CVLR_tauemumu']
        wceff['CVRL'] = wc['CVLR_mumutaue']
        wceff['CVRR'] = wc['CVRR_muetaumu']
        wceff['CSRR'] = wc['CSRR_tauemumu'] - wc['CSRR_muetaumu'] / 2
        wceff['CSLL'] = wc['CSRR_mumuetau'].conjugate() - wc['CSRR_emumutau'].conjugate() / 2
        wceff['CSLR'] = -2 * wc['CVLR_taumumue']
        wceff['CSRL'] = -2 * wc['CVLR_muetaumu']
        wceff['CTLL'] = -wc['CSRR_emumutau'].conjugate() / 8
        wceff['CTRR'] = -wc['CSRR_muetaumu'] / 8
    elif (l0, l1, l2, l3) == ('mu', 'e', 'e', 'e'):
        wceff['C7'] = e / ml0 * wc['Cgamma_mue']
        wceff['C7p'] = e / ml0 * wc['Cgamma_emu'].conjugate()
        wceff['CVLL'] = wc['CVLL_eemue']
        wceff['CVLR'] = wc['CVLR_mueee']
        wceff['CVRL'] = wc['CVLR_eemue']
        wceff['CVRR'] = wc['CVRR_eemue']
        wceff['CSRR'] = wc['CSRR_eemue']
        wceff['CSLL'] = wc['CSRR_eeemu'].conjugate()
    elif (l0, l1, l2, l3) == ('tau', 'e', 'mu', 'e'):
        wceff['CVLL'] = wc['CVLL_muetaue']
        wceff['CVLR'] = wc['CVLR_tauemue']
        wceff['CVRL'] = wc['CVLR_muetaue']
        wceff['CVRR'] = wc['CVRR_muetaue']
        wceff['CSRR'] = wc['CSRR_muetaue']
        wceff['CSLL'] = wc['CSRR_emuetau'].conjugate()
    elif (l0, l1, l2, l3) == ('tau', 'mu', 'e', 'mu'):
        wceff['CVLL'] = wc['CVLL_muemutau'].conjugate()
        wceff['CVLR'] = wc['CVLR_taumuemu']
        wceff['CVRL'] = wc['CVLR_muemutau'].conjugate()
        wceff['CVRR'] = wc['CVRR_muemutau'].conjugate()
        wceff['CSRR'] = wc['CSRR_muemutau'].conjugate()
        wceff['CSLL'] = wc['CSRR_emutaumu']
    else:
        raise ValueError("Decay {}-->{}-{}+{}- not implemented".format(l0, l1, l2, l3))
    return wceff


def BR_taul1l2l3(wc_obj, par, l1, l2, l3):
    r"""Branching ratio of $\tau^-\to\ell_1^-\ell_2^+\ell_3^-$."""
    scale = flavio.config['renormalization scale']['taudecays']
    # cf. (22, 23) of hep-ph/0404211
    wceff = wc_eff(wc_obj, par, scale, 'tau', l1, l2, l3, nf_out=4)
    if (l1, l2, l3) == ('mu', 'e', 'e'):
        br_wc = _BR_taumuee(par['m_tau'], par['m_e'], wceff)
    elif (l1, l2, l3) == ('e', 'mu', 'mu'):
        br_wc = _BR_taumuee(par['m_tau'], par['m_mu'], wceff)
    elif (l1, l2, l3) == ('mu', 'mu', 'mu'):
        br_wc = _BR_tau3mu(par['m_tau'], par['m_mu'], wceff)
    elif (l1, l2, l3) == ('e', 'e', 'e'):
        br_wc = _BR_tau3mu(par['m_tau'], par['m_e'], wceff)
    elif (l1, l2, l3) == ('e', 'mu', 'e'):
        br_wc = _BR_taumuemu(wceff)
    elif (l1, l2, l3) == ('mu', 'e', 'mu'):
        br_wc = _BR_taumuemu(wceff)
    pre_br = par['tau_tau'] * par['m_tau']**5 / 192 / 8 / pi**3
    return pre_br * br_wc


# function returning function needed for prediction instance
def br_taul1l2l3_fct(l1, l2, l3):
    def f(wc_obj, par):
        return BR_taul1l2l3(wc_obj, par, l1, l2, l3)
    return f


# Observable and Prediction instances
_tex = {'e': 'e', 'mu': r'\mu'}

for (l1, l2, l3) in [('mu', 'e', 'e'), ('mu', 'mu', 'mu'),
                     ('e', 'e', 'e'), ('e', 'mu', 'mu'),
                     ('e', 'mu', 'e'), ('mu', 'e', 'mu')]:
    _process_tex = r"\tau^-\to " + _tex[l1] + r"^-" + _tex[l2] + r"^+" + _tex[l3] + r"^-"
    _process_taxonomy = r'Process :: $\tau$ lepton decays :: LFV decays :: $\tau\to \ell^\prime\ell\ell$ :: $' + _process_tex + r"$"

    _obs_name = "BR(tau->" + l1 + l2 + l3 + ")"
    _obs = flavio.classes.Observable(_obs_name)
    _obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    flavio.classes.Prediction(_obs_name, br_taul1l2l3_fct(l1, l2, l3))



import unittest
import numpy as np
import flavio

par = flavio.default_parameters.get_central_all()
wc_obj = flavio.WilsonCoefficients()
wc = flavio.physics.bdecays.wilsoncoefficients.wctot_dict(wc_obj, 'bsmumu', flavio.config['renormalization scale']['bvll'], par)

class TestQCDFInterpolate(unittest.TestCase):
    def test_qcdf_interpolate(self):
        q2 = 3.33333
        B = 'B0'
        V = 'K*0'
        lep = 'mu'
        amps_in = flavio.physics.bdecays.bvll.qcdf_interpolate.helicity_amps_qcdf(q2, par, B, V)
        amps_ex = flavio.physics.bdecays.bvll.qcdf.helicity_amps_qcdf(q2, wc, par, B, V)
        for i in amps_in.keys():
            if not amps_ex[i] == 0:
                self.assertAlmostEqual(amps_in[i]/(amps_ex[i]), 1, places=2)
            else:
                self.assertEqual(amps_in[i], 0)

        # and the same game again for different q2 and process
        q2 = 0.15
        B = 'B+'
        V = 'K*+'
        lep = 'e'
        amps_in = flavio.physics.bdecays.bvll.qcdf_interpolate.helicity_amps_qcdf(q2, par, B, V)
        amps_ex = flavio.physics.bdecays.bvll.qcdf.helicity_amps_qcdf(q2, wc, par, B, V)
        for i in amps_in.keys():
            if not amps_ex[i] == 0:
                self.assertAlmostEqual(amps_in[i]/(amps_ex[i]), 1, places=0)
            else:
                self.assertEqual(amps_in[i], 0)

        # and the same game again for different q2 and process
        q2 = 5.935
        B = 'Bs'
        V = 'phi'
        lep = 'mu'
        amps_in = flavio.physics.bdecays.bvll.qcdf_interpolate.helicity_amps_qcdf(q2, par, B, V)
        amps_ex = flavio.physics.bdecays.bvll.qcdf.helicity_amps_qcdf(q2, wc, par, B, V)
        for i in amps_in.keys():
            if not amps_ex[i] == 0:
                self.assertAlmostEqual(amps_in[i]/(amps_ex[i]), 1, places=0)
            else:
                self.assertEqual(amps_in[i], 0)

    def test_qcdf_contributions(self):
        # check that the individual contributions sum to the total contribution
        q2 = 4.679
        B = 'B0'
        V = 'K*0'
        lep = 'mu'
        amps_all = flavio.physics.bdecays.bvll.qcdf_interpolate.helicity_amps_qcdf(q2, par, B, V)
        amps_WA = flavio.physics.bdecays.bvll.qcdf_interpolate.helicity_amps_qcdf(q2, par, B, V, contribution='WA')
        amps_O8 = flavio.physics.bdecays.bvll.qcdf_interpolate.helicity_amps_qcdf(q2, par, B, V, contribution='O8')
        amps_QSS = flavio.physics.bdecays.bvll.qcdf_interpolate.helicity_amps_qcdf(q2, par, B, V, contribution='QSS')
        for i in amps_all.keys():
            if not amps_all[i] == 0:
                self.assertAlmostEqual(amps_all[i]/(amps_WA[i]+amps_O8[i]+amps_QSS[i]), 1, places=5)



import unittest
import numpy as np
import flavio
import math

s = 1.519267515435317e+24
ps = 1e-12*s


class TestIntegrate(unittest.TestCase):
    def test_nintegrate(self):
        # integrate sin(x) from 0 to 2
        xmin = 0
        xmax = 2
        val = 2*math.sin(1)**2
        self.assertAlmostEqual(flavio.math.integrate.nintegrate(math.sin, xmin, xmax), val, delta=0.01*val)
        self.assertAlmostEqual(flavio.math.integrate.nintegrate_fast(math.sin, xmin, xmax), val, delta=0.01*val)



import unittest
import flavio
from .kll import br_kll
from flavio.physics import ckm
from math import pi


constraints = flavio.default_parameters
wc_obj = flavio.WilsonCoefficients()
par = constraints.get_central_all()


class TestKll(unittest.TestCase):
    def test_klmm_sd_sm(self):
        # compare to hep-ph/0605203
        # correct for different CKM choice
        par06 = par.copy()
        par06.update({'Vus': 0.22715, 'Vub': 0.003683, 'Vcb': 0.04161,
                      'delta': 59.0 / 180 * pi})
        self.assertAlmostEqual(br_kll(par06, wc_obj, 'KL', 'mu', 'mu', ld=False),
                               0.79 * 1e-9,
                               delta=0.02e-9)

    def test_ksmm_sd_sm(self):
        # compare to 1707.06999
        # correct for different CKM choice
        my_xi_t = ckm.xi('t', 'sd')(par)
        xi_t = ckm.xi('t', 'sd')({'Vus': 0.22508, 'Vub': 0.003715, 'Vcb': 0.04181,
                                  'delta': 65.4 / 180 * pi})
        r = (my_xi_t.imag / xi_t.imag)**2
        self.assertAlmostEqual(br_kll(par, wc_obj, 'KS', 'mu', 'mu', ld=False),
                               r * 0.19e-12,
                               delta=0.02e-12)

    def test_ksmm_ld_sm(self):
        # 1712.01295 eq. (2.33)
        _par = par.copy()
        _par.update({'Vus': 0.22508, 'Vub': 0.003715, 'Vcb': 0.04181,
                      'delta': 65.4 / 180 * pi})
        self.assertAlmostEqual(br_kll(_par, wc_obj, 'KS', 'mu', 'mu', ld=True)
                               - br_kll(_par, wc_obj, 'KS', 'mu', 'mu', ld=False),
                               4.99e-12,
                               delta=0.1e-12)

    def test_klmm_sm_p(self):
        par_p = par.copy()
        par_p['chi_disp(KL->gammagamma)'] = 0.71
        par_p.update({'Vus': 0.22508, 'Vub': 0.003715, 'Vcb': 0.04181,
                      'delta': 65.4 / 180 * pi})
        self.assertAlmostEqual(br_kll(par_p, wc_obj, 'KL', 'mu', 'mu', ld=True),
                               6.85e-9,
                               delta=0.2e-9)

    def test_klmm_sm_m(self):
        par_m = par.copy()
        par_m['chi_disp(KL->gammagamma)'] = -0.71
        par_m.update({'Vus': 0.22508, 'Vub': 0.003715, 'Vcb': 0.04181,
                      'delta': 65.4 / 180 * pi})
        self.assertAlmostEqual(br_kll(par_m, wc_obj, 'KL', 'mu', 'mu', ld=True),
                               8.11e-9,
                               delta=0.2e-9)

    def test_ksmm_sm(self):
        # 1712.01295 eq. (2.33)
        self.assertAlmostEqual(flavio.sm_prediction('BR(KS->mumu)'),
                               5.2e-12,
                               delta=0.2e-12)



import unittest
import flavio


class TestTauVl(unittest.TestCase):
    def test_tauvl(self):
        # compare to the experimental values
        self.assertEqual(flavio.sm_prediction('BR(tau->rhoe)'), 0)
        self.assertEqual(flavio.sm_prediction('BR(tau->rhomu)'), 0)



"""Common functions for physics."""

from collections import Counter

def conjugate_par(par_dict):
    """Given a dictionary of parameter values, return the dictionary where
    all CP-odd parameters have flipped sign.

    This assumes that the only CP-odd parameters are `gamma` or `delta` (the
    CKM phase in the Wolfenstein or standard parametrization)."""
    cp_odd = ['gamma', 'delta']
    return {k: -v if k in cp_odd else v for k, v in par_dict.items()}

def conjugate_wc(wc_dict):
    """Given a dictionary of Wilson coefficients, return the dictionary where
    all coefficients are CP conjugated (which simply amounts to complex
    conjugation)."""
    return {k: v.conjugate() for k, v in wc_dict.items()}

def add_dict(dicts):
    """Add dictionaries.

    This will add the two dictionaries

    `A = {'a':1, 'b':2, 'c':3}`
    `B = {'b':3, 'c':4, 'd':5}`

    into

    `A + B {'a':1, 'b':5, 'c':7, 'd':5}`

    but works for an arbitrary number of dictionaries."""
    # start with the first dict
    res = Counter(dicts[0])
    # successively add all other dicts
    for d in dicts[1:]: res.update(Counter(d))
    return res


def lambda_K(a, b, c):
    r"""Källén function $\lambda$.

    $\lambda(a,b,c) = a^2 + b^2 + c^2 - 2 (ab + bc + ac)$
    """
    z = a**2 + b**2 + c**2 - 2 * (a * b + b * c + a * c)
    if z < 0:
        # to avoid sqrt(-1e-16) type errors due to numerical inaccuracies
        return 0
    else:
        return z



r"""Functions for the anomalous magnetic moments of leptons"""

import flavio
from math import sqrt, pi


def al(wc_obj, par, l, eft, scale):
    r"""Anomalous magnetic moment of lepton `l` at scale `scale`,
    $a_\ell = (g_\ell - 2) / 2$."""
    wc = wc_obj.get_wc(sector='dF=0', scale=scale, par=par, eft=eft, basis='flavio')
    # the covariant derivative contains d_mu + eta i e Q A, where Q=-1
    # the Lagrangian contains -eta * e*Q*a/(4*m) (...)
    # the effective Lagrangian contains p * ReC7 (...)
    # with p = 4*GF/sqrt(2)*e/(16*pi**2)*m
    # thus a = -eta*4*m*p/e/Q * ReC7 = pre * ReC7
    GF = par['GF']
    m = par['m_' + l]
    pre = m**2 * GF / sqrt(2) / pi**2
    C7 = wc['C7_' + 2 * l]  # e.g. C7_mumu
    aSM = par['a_{} SM'.format(l)]  # e.g. a_mu SM
    return aSM + pre * C7.real


def ae(wc_obj, par):
    r"""Anomalous magnetic moment of the electron,
    $a_e = (g_e - 2) / 2$."""
    scale = flavio.config['renormalization scale']['e g-2']
    return al(wc_obj, par, l='e', eft='WET-3', scale=scale)

def amu(wc_obj, par):
    r"""Anomalous magnetic moment of the muon,
    $a_\mu = (g_\mu - 2) / 2$."""
    scale = flavio.config['renormalization scale']['mudecays']
    return al(wc_obj, par, l='mu', eft='WET-3', scale=scale)

def atau(wc_obj, par):
    r"""Anomalous magnetic moment of the tau lepton,
    $a_\tau = (g_\tau - 2) / 2$."""
    scale = flavio.config['renormalization scale']['tau g-2']
    return al(wc_obj, par, l='tau', eft='WET', scale=scale)


def make_observable(l, ltex, lname, lfunc):
    """Instantiate the observable and add metadata."""
    obs_name = "a_" + l
    obs = flavio.classes.Observable(obs_name)
    obs.set_description(r"Anomalous magnetic moment of the " + lname)
    obs.tex = r"$a_{}$".format(ltex)
    obs.add_taxonomy(r"Process :: Dipole moments :: Lepton anomalous magnetic moments :: $a_{}$".format(ltex))
    flavio.classes.Prediction(obs_name, lfunc)


make_observable('mu', r'\mu', 'muon', amu)
make_observable('tau', r'\tau', 'tau lepton', atau)
make_observable('e', r'e', 'electron', ae)



r"""Functions for lepton flavour violating $\tau\to \ell\gamma$ decays."""

import flavio
from flavio.physics.taudecays.common import GammaFvf


# names of LFV sectors in WCxf
wcxf_sector_names = {('tau', 'mu'): 'mutau',
                     ('tau', 'e'): 'taue',
                     ('mu', 'e'): 'mue', }


def BR_llgamma(wc_obj, par, l1, l2):
    r"""Branching ratio of $\ell_1\to \ell_2\gamma$."""
    scale = flavio.config['renormalization scale'][l1 + 'decays']
    ll = wcxf_sector_names[l1, l2]
    wc = wc_obj.get_wc(ll, scale, par, nf_out=4)
    ml1 = par['m_' + l1]
    ml2 = par['m_' + l2]
    gTL = 2 * wc['Cgamma_' + l1 + l2].conjugate()
    gTR = 2 * wc['Cgamma_' + l2 + l1]
    return (par['tau_' + l1] * GammaFvf(M=ml1, mv=0, mf=ml2,
                                        gL=0, gR=0,
                                        gTL=gTL, gtTL=0,
                                        gTR=gTR, gtTR=0))


def BR_taulgamma(wc_obj, par, lep):
    r"""Branching ratio of $\tau\to \ell\gamma$."""
    return BR_llgamma(wc_obj, par, 'tau', lep)


# function returning function needed for prediction instance
def br_taulgamma_fct(lep):
    def f(wc_obj, par):
        return BR_taulgamma(wc_obj, par, lep)
    return f


# Observable and Prediction instances
_tex = {'e': 'e', 'mu': r'\mu'}

for lep in _tex:
    _process_tex = r"\tau\to " + _tex[lep] + r"\gamma"
    _process_taxonomy = r'Process :: $\tau$ lepton decays :: LFV decays :: $\tau\to \ell\gamma$ :: $' + _process_tex + r"$"

    _obs_name = "BR(tau->" + lep + "gamma)"
    _obs = flavio.classes.Observable(_obs_name)
    _obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    flavio.classes.Prediction(_obs_name, br_taulgamma_fct(lep))



import unittest
import numpy as np
from .bvlnu import *
from flavio.physics.bdecays.formfactors.b_v import bsz_parameters
from flavio.physics.eft import WilsonCoefficients
from flavio.parameters import default_parameters
import flavio
from math import pi

constraints = default_parameters
wc_obj = WilsonCoefficients()
par = constraints.get_central_all()

class TestBVll(unittest.TestCase):
    def test_brhoee(self):
        q2 = 3.5
        self.assertEqual(
            dBRdq2(q2, wc_obj, par, 'B0', 'rho+', 'e', A=None),
            flavio.Observable["dBR/dq2(B0->rhoenu)"].prediction_central(constraints, wc_obj, q2=q2) )

    def test_decays(self):
        # just check if any of the modes raises an exception
        flavio.Observable["dBR/dq2(B0->rhoenu)"].prediction_central(constraints, wc_obj, q2=3)
        flavio.Observable["dBR/dq2(B+->rhoenu)"].prediction_central(constraints, wc_obj, q2=3)
        flavio.Observable["dBR/dq2(B0->D*enu)"].prediction_central(constraints, wc_obj, q2=3)
        flavio.Observable["dBR/dq2(B+->D*enu)"].prediction_central(constraints, wc_obj, q2=3)
        flavio.Observable["dBR/dq2(B+->omegaenu)"].prediction_central(constraints, wc_obj, q2=3)
        flavio.Observable["dBR/dq2(Bs->K*enu)"].prediction_central(constraints, wc_obj, q2=3)
        flavio.Observable["dBR/dcl(B0->rhoenu)"].prediction_central(constraints, wc_obj, cl=0.5)
        flavio.Observable["dBR/dcl(B+->rhoenu)"].prediction_central(constraints, wc_obj, cl=0.5)
        flavio.Observable["dBR/dcl(B0->D*enu)"].prediction_central(constraints, wc_obj, cl=0.5)
        flavio.Observable["dBR/dcl(B+->D*enu)"].prediction_central(constraints, wc_obj, cl=0.5)
        flavio.Observable["dBR/dcl(B+->omegaenu)"].prediction_central(constraints, wc_obj, cl=0.5)
        flavio.Observable["dBR/dcl(Bs->K*enu)"].prediction_central(constraints, wc_obj, cl=0.5)
        flavio.Observable["dBR/dcV(B0->rhoenu)"].prediction_central(constraints, wc_obj, cV=0.5)
        flavio.Observable["dBR/dcV(B+->rhoenu)"].prediction_central(constraints, wc_obj, cV=0.5)
        flavio.Observable["dBR/dcV(B0->D*enu)"].prediction_central(constraints, wc_obj, cV=0.5)
        flavio.Observable["dBR/dcV(B+->D*enu)"].prediction_central(constraints, wc_obj, cV=0.5)
        flavio.Observable["dBR/dcV(B+->omegaenu)"].prediction_central(constraints, wc_obj, cV=0.5)
        flavio.Observable["dBR/dcV(Bs->K*enu)"].prediction_central(constraints, wc_obj, cV=0.5)
        flavio.Observable["dBR/dphi(B0->rhoenu)"].prediction_central(constraints, wc_obj, phi=1.5)
        flavio.Observable["dBR/dphi(B+->rhoenu)"].prediction_central(constraints, wc_obj, phi=1.5)
        flavio.Observable["dBR/dphi(B0->D*enu)"].prediction_central(constraints, wc_obj, phi=1.5)
        flavio.Observable["dBR/dphi(B+->D*enu)"].prediction_central(constraints, wc_obj, phi=1.5)
        flavio.Observable["dBR/dphi(B+->omegaenu)"].prediction_central(constraints, wc_obj, phi=1.5)
        flavio.Observable["dBR/dphi(Bs->K*enu)"].prediction_central(constraints, wc_obj, phi=1.5)

    def test_binned(self):
        # this is all the total BR, calulated in 4 different ways
        self.assertAlmostEqual(
        flavio.Observable["BR(B0->D*enu)"].prediction_central(
                                constraints, wc_obj),
        flavio.Observable["<BR>/<cl>(B0->D*enu)"].prediction_central(
                                constraints, wc_obj, clmin=-1, clmax=1))
        self.assertAlmostEqual(
        flavio.Observable["BR(B0->D*enu)"].prediction_central(
                                constraints, wc_obj),
        flavio.Observable["<BR>/<cV>(B0->D*enu)"].prediction_central(
                                constraints, wc_obj, cVmin=-1, cVmax=1))
        self.assertAlmostEqual(
        flavio.Observable["BR(B0->D*enu)"].prediction_central(
                                constraints, wc_obj),
        flavio.Observable["<BR>/<phi>(B0->D*enu)"].prediction_central(
                                constraints, wc_obj, phimin=-pi, phimax=pi))

    def test_bvlnu_nu(self):
        wc_sm = flavio.WilsonCoefficients()
        wc_np_tau = flavio.WilsonCoefficients()
        wc_np_tau.set_initial({'CVL_butaunutau': 1}, 4.8)
        wc_np_e = flavio.WilsonCoefficients()
        wc_np_e.set_initial({'CVL_butaunue': 1}, 4.8)
        obs = flavio.Observable["BR(B+->rhotaunu)"]
        br_sm = obs.prediction_central(constraints, wc_sm)
        br_tau = obs.prediction_central(constraints, wc_np_tau)
        br_e = obs.prediction_central(constraints, wc_np_e)
        # with interference: (1 + 1)^2 = 4
        self.assertAlmostEqual(br_tau/br_sm, 4, delta=0.04)
        # without interference: 1 + 1 = 2
        self.assertAlmostEqual(br_e/br_sm, 2, delta=0.02)

    def test_BRLT(self):
        self.assertAlmostEqual(flavio.sm_prediction('dBR/dq2(B0->D*enu)', q2=3),
                               flavio.sm_prediction('dBR_L/dq2(B0->D*enu)', q2=3) +
                               flavio.sm_prediction('dBR_T/dq2(B0->D*enu)', q2=3),
                               places=10)

    def test_bdstau_binned(self):
        # for the full kinematical range, it should integrate to 1
        self.assertAlmostEqual(
            flavio.sm_prediction('<BR>/BR(B->D*taunu)', 3.1, 11.7), 1, delta=0.03)

    def test_FL(self):
        self.assertAlmostEqual(flavio.sm_prediction('<FL>(B0->D*taunu)', 3.15, 10.71),
                               0.46, delta=0.022)
        self.assertAlmostEqual(flavio.sm_prediction('FLtot(B0->D*taunu)'),
                               flavio.sm_prediction('<FL>(B0->D*taunu)', 3.15, 10.71),
                               delta=0.001)



import unittest
import numpy as np
import flavio

constraints = flavio.default_parameters
wc_obj = flavio.WilsonCoefficients()
par = constraints.get_central_all()

class TestTauPnu(unittest.TestCase):
    def test_taupnu(self):
        # compare to the experimental values
        self.assertAlmostEqual(
            flavio.Observable.get_instance("BR(tau->pinu)").prediction_central(constraints, wc_obj),
            10.82e-2, delta=2*0.05e-2)
        self.assertAlmostEqual(
            flavio.Observable.get_instance("BR(tau->Knu)").prediction_central(constraints, wc_obj),
            6.96e-3, delta=2*0.1e-3)



r"""Functions for radiative, leptonic, and semi-leptonic tree-level as well
as FCNC $B$ decays."""

from . import formfactors
from . import bvll
from . import bvlnu
from . import bll
from . import blnu
from . import bpll
from . import bpll_subleading
from . import bpll_lfv
from . import bplnu
from . import bxlnu
from . import bvnunu
from . import bpnunu
from . import bvgamma
from . import bxgamma
from . import lambdablambdall
from . import lambdablambdall_subleading
from . import bxll
from . import bc_lifetime



r"""Observables in lepton flavour violating decays $B\to V\ell\ell^\prime$"""

import flavio
import scipy.integrate
from flavio.classes import Observable, Prediction
from math import sqrt, pi


def prefactor(q2, par, B, V, l1, l2):
    GF = par['GF']
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    scale = flavio.config['renormalization scale']['bvll']
    alphaem = flavio.physics.running.running.get_alpha(par, scale)['alpha_e']
    di_dj = flavio.physics.bdecays.common.meson_quark[(B,V)]
    xi_t = flavio.physics.ckm.xi('t',di_dj)(par)
    if q2 <= (ml1+ml2)**2:
        return 0
    return 4*GF/sqrt(2)*xi_t*alphaem/(4*pi)

def helicity_amps(q2, wc, par, B, V, l1, l2):
    scale = flavio.config['renormalization scale']['bvll']
    wc_eff = flavio.physics.bdecays.wilsoncoefficients.get_wceff_lfv(q2, wc, par, B, V, l1, l2, scale)
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    mB = par['m_'+B]
    mV = par['m_'+V]
    mb = flavio.physics.running.running.get_mb(par, scale)
    N = prefactor(q2, par, B, V, l1, l2)
    ff = flavio.physics.bdecays.bvll.amplitudes.get_ff(q2, par, B, V)
    h = flavio.physics.bdecays.angular.helicity_amps_v(q2, mB, mV, mb, 0, ml1, ml2, ff, wc_eff, N)
    return h

def bvlilj_obs(function, q2, wc, par, B, V, l1, l2):
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    mB = par['m_'+B]
    mV = par['m_'+V]
    if q2 < (ml1+ml2)**2 or q2 > (mB-mV)**2:
        return 0
    scale = flavio.config['renormalization scale']['bvll']
    mb = flavio.physics.running.running.get_mb(par, scale)
    h = helicity_amps(q2, wc, par, B, V, l1, l2)
    J = flavio.physics.bdecays.angular.angularcoeffs_general_v(h, q2, mB, mV, mb, 0, ml1, ml2)
    return function(J)

def bvlilj_obs_int(function, q2min, q2max, wc, par, B, V, l1, l2):
    def obs(q2):
        return bvlilj_obs(function, q2, wc, par, B, V, l1, l2)
    return flavio.math.integrate.nintegrate(obs, q2min, q2max)

def BR_tot(wc_obj, par, B, V, l1, l2):
    scale = flavio.config['renormalization scale']['bvll']
    label = flavio.physics.bdecays.common.meson_quark[(B,V)] + l1 + l2 # e.g. bsemu, bdtaue
    wc = wc_obj.get_wc(label, scale, par)
    if all([abs(v) < 1e-12 for v in wc.values()]):
        # if all WCs are essentially zero, return BR=0
        return 0
    mB = par['m_'+B]
    mV = par['m_'+V]
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    q2max = (mB-mV)**2
    q2min = (ml1+ml2)**2
    tauB = par['tau_'+B]
    return tauB*bvlilj_obs_int(flavio.physics.bdecays.bvll.observables.dGdq2,
                          q2min, q2max, wc, par, B, V, l1, l2)

def BR_tot_function(B, V, l1, l2):
    return lambda wc_obj, par: BR_tot(wc_obj, par, B, V, l1, l2)

def BR_tot_function_leptonsum(B, V, l1, l2):
    return lambda wc_obj, par: BR_tot(wc_obj, par, B, V, l1, l2) + BR_tot(wc_obj, par, B, V, l2, l1)



# Observable and Prediction instances

_tex = {'emu': r'e^+\mu^-', 'mue': r'\mu^+e^-',
    'taue': r'\tau^+e^-', 'etau': r'e^+\tau^-',
    'taumu': r'\tau^+\mu^-', 'mutau': r'\mu^+\tau^-'}
_tex_lsum = {'emu': r'e^\pm\mu^\mp',  'etau': r'e^\pm\tau^\mp', 'mutau': r'\mu^\pm\tau^\mp'}
_func = {'BR': BR_tot_function, }
_desc = { 'BR': 'Total', }
_tex_br = {'BR': r'\text{BR}', }
_args = {'BR': None, }
_hadr = {
'B0->K*': {'tex': r"\bar B^0\to \bar K^{*0}", 'B': 'B0', 'V': 'K*0', },
'B+->K*': {'tex': r"B^-\to K^{*-}", 'B': 'B+', 'V': 'K*+', },
'B+->rho': {'tex': r"B^-\to \rho^{-}", 'B': 'B+', 'V': 'rho+', },
'B0->rho': {'tex': r"\bar B^0\to \rho^{0}", 'B': 'B0', 'V': 'rho0', },
'Bs->phi': {'tex': r"\bar B_s\to \phi", 'B': 'Bs', 'V': 'phi', },
}

for ll in [('e','mu'), ('mu','e'), ('e','tau'), ('tau','e'), ('mu','tau'), ('tau','mu')]:
    for M in _hadr.keys():

        _process_tex = _hadr[M]['tex']+' '+_tex[''.join(ll)]
        _process_taxonomy = r'Process :: $b$ hadron decays :: FCNC decays :: $B\to V\ell^+\ell^-$ :: $' + _process_tex + r"$"

        for br in ['BR',]:
            _obs_name = br + "("+M+''.join(ll)+")"
            _obs = Observable(_obs_name)
            _obs.set_description(_desc[br] + r" branching ratio of $"+_process_tex+r"$")
            _obs.tex = r'$' + _tex_br[br] + "(" + _process_tex+r")$"
            _obs.arguments = _args[br]
            _obs.add_taxonomy(_process_taxonomy)
            Prediction(_obs_name, _func[br](_hadr[M]['B'], _hadr[M]['V'], ll[0], ll[1]))



"""Functions for working with YAML files."""

import yaml
from collections import OrderedDict
import os


def represent_dict_order(self, data):
    return self.represent_mapping('tag:yaml.org,2002:map', data.items())
yaml.add_representer(OrderedDict, represent_dict_order)


class SafeIncludeLoader(yaml.SafeLoader):
    """PyYAML loader supporting the `!include` and `!include_merge_list`
    constructors."""
    def __init__(self, stream):
        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            # this happens when `stream` is a string
            self._root = None
        super().__init__(stream)

    def include(self, node):
        filename = os.path.join(self._root, self.construct_scalar(node))
        with open(filename, 'r') as f:
            return yaml.load(f, self.__class__)

    def include_merge_list(self, node):
        files = self.construct_sequence(node)
        a = []
        for file in files:
            try:
                filename = os.path.join(self._root, file)
                with open(filename, 'r') as f:
                    a += yaml.load(f, self.__class__)
            except (FileNotFoundError, TypeError):
                a.append(file)
        return a


SafeIncludeLoader.add_constructor('!include',
                                  SafeIncludeLoader.include)
SafeIncludeLoader.add_constructor('!include_merge_list',
                                  SafeIncludeLoader.include_merge_list)

def load_include(*args, **kwargs):
    return yaml.load(*args, Loader=SafeIncludeLoader, **kwargs)



r"""Functions for $\tau\to \ell \nu\nu$ decays."""

import flavio
from math import log, sqrt, pi


def F(x):
    return 1 - 8*x + 8*x**3 - x**4 - 12*x**2*log(x)

def G(x):
    return 1 + 9*x - 9*x**2 - x**3 + 6*x*log(x) + 6*x**2*log(x)

def _BR(x, CL, CR):
    return F(x) * (abs(CL)**2 + abs(CR)**2) - 4 * G(x) * (CL * CR.conjugate()).real


def GFeff(wc_obj, par):
    r"""Effective Fermi constant in the presence of new physics."""
    scale = flavio.config['renormalization scale']['mudecays']
    wc = wc_obj.get_wc('nunumue', scale, par, eft='WET-3')
    CL = wc['CVLL_numunueemu']
    CR = wc['CVLR_numunueemu']
    me = par['m_e']
    mmu = par['m_mu']
    GF = par['GF']
    x = me**2 / mmu**2
    CLSM = -4 *  GF / sqrt(2)
    r = _BR(x, CL + CLSM, CR) / _BR(x, CLSM, 0)
    return GF / sqrt(r)


def BR_taulnunu(wc_obj, par, lep, nu1, nu2):
    r"""BR of $\tau\to l nu_1\bar nu_2$ for specific neutrino flavours"""
    if lep == 'e':
        sec = 'nunutaue'
    elif lep == 'mu':
        sec = 'nunumutau'
    scale = flavio.config['renormalization scale']['taudecays']
    wc = wc_obj.get_wc(sec, scale, par, eft='WET-4')
    ml = par['m_' + lep]
    mtau = par['m_tau']
    x = ml**2 / mtau**2
    nnll = 'nu{}nu{}tau{}'.format(nu2, nu1, lep)
    try:
        CL = wc['CVLL_' + nnll]
        CR = wc['CVLR_' + nnll]
    except KeyError:
        nnll = 'nu{}nu{}{}tau'.format(nu1, nu2, lep)
        CL = wc['CVLL_' + nnll].conjugate()
        CR = wc['CVLR_' + nnll].conjugate()
    if nu1 == 'tau' and nu2 == lep:
        # SM contribution, taking into account NP in mu->enunu!
        GF = GFeff(wc_obj, par)
        CL += -4 * GF / sqrt(2)
    pre = par['tau_tau'] / 3 / 2**9 / pi**3 * mtau**5
    alpha_e = flavio.physics.running.running.get_alpha_e(par, scale, nf_out=4)
    # eq. (3) of arXiv:1310.7922
    emcorr = 1 + alpha_e / (2 * pi) * (25 / 4 - pi**2)
    return pre * _BR(x, CL, CR) * emcorr


def BR_taulnunu_summed(wc_obj, par, lep):
    """BR of tau->lnunu summed over neutrino flavours"""
    _l =  ['e', 'mu', 'tau']
    return sum([BR_taulnunu(wc_obj, par, lep, nu1, nu2) for nu1 in _l for nu2 in _l])


# function returning function needed for prediction instance
def br_taulnunu(lep):
    def f(wc_obj, par):
        return BR_taulnunu_summed(wc_obj, par, lep)
    return f


# Observable and Prediction instances
_tex = {'e': 'e', 'mu': r'\mu'}

for lep in _tex:
    _process_tex = r"\tau^-\to " + _tex[lep] + r"^- \nu\bar\nu"
    _process_taxonomy = r'Process :: $\tau$ lepton decays :: Leptonic tree-level decays :: $\tau\to \ell\nu\bar\nu$ :: $' + _process_tex + r"$"

    _obs_name = "BR(tau->" + lep + "nunu)"
    _obs = flavio.classes.Observable(_obs_name)
    _obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
    _obs.add_taxonomy(_process_taxonomy)
    flavio.classes.Prediction(_obs_name, br_taulnunu(lep))



r"""Functions for $\tau\to P\nu$."""

import flavio
from flavio.physics.taudecays import common
from flavio.physics.bdecays.wilsoncoefficients import get_CVLSM
from math import sqrt
from flavio.physics.taudecays.taulnunu import GFeff


def _br_taupnu(wc_obj, par, P, lep):
    r"""Branching ratio of $\tau^+\to P^+\bar\nu_\ell$."""
    # CKM element
    scale = flavio.config['renormalization scale']['taudecays']
    if P == 'pi+':
        Vij = flavio.physics.ckm.get_ckm(par)[0, 0]  # Vud
        qqlnu = 'dutaunu' + lep
        mq1 = flavio.physics.running.running.get_md(par, scale)
        mq2 = flavio.physics.running.running.get_mu(par, scale)
    elif P == 'K+':
        Vij = flavio.physics.ckm.get_ckm(par)[0, 1]  # Vus
        qqlnu = 'sutaunu' + lep
        mq1 = flavio.physics.running.running.get_ms(par, scale)
        mq2 = flavio.physics.running.running.get_mu(par, scale)
    # Wilson coefficients
    wc = wc_obj.get_wc(qqlnu, scale, par, nf_out=4)
    # add SM contribution to Wilson coefficient
    if lep == 'tau':
        # for the SM contribution, need the Fermi constant with possible
        # NP effects in mu->enunu subtracted, not the measured one
        r_GF = GFeff(wc_obj, par) / par['GF']
        wc['CVL_' + qqlnu] += get_CVLSM(par, scale, nf=4) * r_GF
    mtau = par['m_tau']
    mP = par['m_' + P]
    rWC = ((wc['CVL_' + qqlnu] - wc['CVR_' + qqlnu])
           + mP**2 / mtau / (mq1 + mq2) * (wc['CSR_' + qqlnu] - wc['CSL_' + qqlnu]))
    gR = -sqrt(2) * par['GF'] * Vij * rWC * par['f_' + P] * par['m_tau']
    return par['tau_tau'] * common.GammaFsf(mtau, mP, 0, 0, gR)


def br_taupnu(wc_obj, par, P):
    r"""Branching ratio of $\tau^+\to P^+\bar\nu_\ell$, summing over the
    neutrino flavour."""
    return sum([_br_taupnu(wc_obj, par, P, lep) for lep in ['e', 'mu', 'tau']])


# function returning function needed for prediction instance
def br_taupnu_fct(P):
    def f(wc_obj, par):
        return br_taupnu(wc_obj, par, P)
    return f

# Observable and Prediction instances

_had = {'pi+': r'\pi^+', 'K+': r'K^+',}
_shortname = {'pi+': 'pi', 'K+': 'K'}

for P in _had:
    _obs_name = "BR(tau->" + _shortname[P]+"nu)"
    _obs = flavio.classes.Observable(_obs_name)
    _process_tex = r"\tau^+\to " + _had[P] + r"\bar\nu"
    _process_taxonomy = r'Process :: $\tau$ lepton decays :: Hadronic tree-level decays :: $\tau\to V\ell$ :: $' + _process_tex + r"$"
    _obs.add_taxonomy(_process_taxonomy)
    _obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
    _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
    flavio.classes.Prediction(_obs_name, br_taupnu_fct(P))



import unittest
from flavio.classes import *
from flavio.statistics.fits import *
from .pypmc import *
from flavio.statistics.probability import NormalDistribution, MultivariateNormalDistribution


class TestPypmcScan(unittest.TestCase):

    def test_pypmc_scan(self):
        # dummy observables
        o1 = Observable( 'test_obs 1' )
        o2 = Observable( 'test_obs 2' )
        # dummy predictions
        def f1(wc_obj, par_dict):
            return par_dict['m_b']
        def f2(wc_obj, par_dict):
            return 2.5
        par = flavio.default_parameters.copy()
        par.set_constraint('m_b', '4.2+-0.2')
        par.set_constraint('m_c', '1.2+-0.1')
        Prediction( 'test_obs 1', f1 )
        Prediction( 'test_obs 2', f2 )
        d1 = NormalDistribution(5, 0.2)
        cov2 = [[0.1**2, 0.5*0.1*0.3], [0.5*0.1*0.3, 0.3**2]]
        d2 = MultivariateNormalDistribution([6,2], cov2)
        m1 = Measurement( 'measurement 1 of test_obs 1' )
        m2 = Measurement( 'measurement 2 of test_obs 1 and test_obs 2' )
        m1.add_constraint(['test_obs 1'], d1)
        m2.add_constraint(['test_obs 1', 'test_obs 2'], d2)
        fit = BayesianFit('fit_emcee_test', par, ['m_b', 'm_c'],  [], ['test_obs 1', 'test_obs 2'])
        scan = pypmcScan(fit)
        scan.run(3, burnin=0)
        self.assertTupleEqual(scan.result.shape, (3, 2))
        BayesianFit.del_instance('fit_emcee_test')
        Observable.del_instance('test_obs 1')
        Observable.del_instance('test_obs 2')
        Measurement.del_instance('measurement 1 of test_obs 1')
        Measurement.del_instance('measurement 2 of test_obs 1 and test_obs 2')



r"""$W$ mass prediction in the presence of new physics."""


from math import log, sqrt
import flavio
from flavio.physics.zdecays import smeftew


def mW_SM(par):
    r"""Two-loop SM prediction for the $W$ pole mass.

    Eq. (6) of arXiv:hep-ph/0311148."""
    dH = log(par['m_h'] / 100)
    dh = (par['m_h'] / 100)**2
    dt = (par['m_t'] / 174.3)**2 - 1
    dZ = par['m_Z'] / 91.1876 - 1
    dalpha = 0  # FIXME
    dalphas = par['alpha_s'] / 0.119 - 1
    m0W = 80.3779
    c1 = 0.05427
    c2 = 0.008931
    c3 = 0.0000882
    c4 = 0.000161
    c5 = 1.070
    c6 = 0.5237
    c7 = 0.0679
    c8 = 0.00179
    c9 = 0.0000664
    c10 = 0.0795
    c11 = 114.9
    return (m0W - c1 * dH - c2 * dH**2 + c3 * dH**4 + c4 * (dh - 1) - c5 * dalpha
            + c6 * dt - c7 * dt**2 - c8 * dH * dt + c9 * dh * dt
            - c10 * dalphas + c11 * dZ)

def dmW_SMEFT(par, C):
    r"""Shift in the $W$ mass due to dimension-6 operators.

    Eq. (2) of arXiv:1606.06502."""
    sh2 = smeftew._sinthetahat2(par)
    sh = sqrt(sh2)
    ch2 = 1 - sh2
    ch = sqrt(ch2)
    Dh = ch * sh / ((ch**2 - sh**2) * 2 * sqrt(2) * par['GF'])
    return Dh * (4 * C['phiWB'] + ch / sh * C['phiD'] + 2 * sh / ch * (C['phil3_11'] + C['phil3_22']) - 2 * sh / ch * C['ll_1221']).real

def mW(wc_obj, par):
    r"""$W$ pole mass."""
    scale = flavio.config['renormalization scale']['wdecays']
    C = wc_obj.get_wcxf(sector='all', scale=scale, par=par,
                        eft='SMEFT', basis='Warsaw')
    dmW = dmW_SMEFT(par, C)
    return mW_SM(par) * (1 - dmW)


_obs_name = "m_W"
_obs = flavio.classes.Observable(_obs_name)
_obs.tex = r"$m_W$"
_obs.set_description(r"$W^\pm$ boson pole mass")
flavio.classes.Prediction(_obs_name, mW)



"""Classes for effective field theory (EFT) Wilson coefficients"""

import wcxf
import wilson


# sector names prior to v0.27 translated to WCxf sector names
sectors_flavio2wcxf = {
 'bcenue': 'cbenu',
 'bcenumu': 'cbenu',
 'bcenutau': 'cbenu',
 'bcmunue': 'cbmunu',
 'bcmunumu': 'cbmunu',
 'bcmunutau': 'cbmunu',
 'bctaunue': 'cbtaunu',
 'bctaunumu': 'cbtaunu',
 'bctaunutau': 'cbtaunu',
 'bdbd': 'dbdb',
 'bdee': 'db',
 'bdemu': 'dbmue',
 'bdetau': 'dbtaue',
 'bdmue': 'dbemu',
 'bdmumu': 'db',
 'bdmutau': 'dbtaumu',
 'bdnuenue': 'dbnunu',
 'bdnuenumu': 'dbnunu',
 'bdnuenutau': 'dbnunu',
 'bdnumunue': 'dbnunu',
 'bdnumunumu': 'dbnunu',
 'bdnumunutau': 'dbnunu',
 'bdnutaunue': 'dbnunu',
 'bdnutaunumu': 'dbnunu',
 'bdnutaunutau': 'dbnunu',
 'bdtaue': 'dbetau',
 'bdtaumu': 'dbmutau',
 'bdtautau': 'db',
 'bsbs': 'sbsb',
 'bsee': 'sb',
 'bsemu': 'sbmue',
 'bsetau': 'sbtaue',
 'bsmue': 'sbemu',
 'bsmumu': 'sb',
 'bsmutau': 'sbtaumu',
 'bsnuenue': 'sbnunu',
 'bsnuenumu': 'sbnunu',
 'bsnuenutau': 'sbnunu',
 'bsnumunue': 'sbnunu',
 'bsnumunumu': 'sbnunu',
 'bsnumunutau': 'sbnunu',
 'bsnutaunue': 'sbnunu',
 'bsnutaunumu': 'sbnunu',
 'bsnutaunutau': 'sbnunu',
 'bstaue': 'sbetau',
 'bstaumu': 'sbmutau',
 'bstautau': 'sb',
 'buenue': 'ubenu',
 'buenumu': 'ubenu',
 'buenutau': 'ubenu',
 'bumunue': 'ubmunu',
 'bumunumu': 'ubmunu',
 'bumunutau': 'ubmunu',
 'butaunue': 'ubtaunu',
 'butaunumu': 'ubtaunu',
 'butaunutau': 'ubtaunu',
 'dcenue': 'cdenu',
 'dcenumu': 'cdenu',
 'dcenutau': 'cdenu',
 'dcmunue': 'cdmunu',
 'dcmunumu': 'cdmunu',
 'dcmunutau': 'cdmunu',
 'dctaunue': 'cdtaunu',
 'dctaunumu': 'cdtaunu',
 'dctaunutau': 'cdtaunu',
 'duenue': 'udenu',
 'duenumu': 'udenu',
 'duenutau': 'udenu',
 'dumunue': 'udmunu',
 'dumunumu': 'udmunu',
 'dumunutau': 'udmunu',
 'dutaunue': 'udtaunu',
 'dutaunumu': 'udtaunu',
 'dutaunutau': 'udtaunu',
 'scenue': 'csenu',
 'scenumu': 'csenu',
 'scenutau': 'csenu',
 'scmunue': 'csmunu',
 'scmunumu': 'csmunu',
 'scmunutau': 'csmunu',
 'sctaunue': 'cstaunu',
 'sctaunumu': 'cstaunu',
 'sctaunutau': 'cstaunu',
 'sdnuenue': 'sdnunu',
 'sdnuenumu': 'sdnunu',
 'sdnuenutau': 'sdnunu',
 'sdnumunue': 'sdnunu',
 'sdnumunumu': 'sdnunu',
 'sdnumunutau': 'sdnunu',
 'sdnutaunue': 'sdnunu',
 'sdnutaunumu': 'sdnunu',
 'sdnutaunutau': 'sdnunu',
 'sdee': 'sd',
 'sdmumu': 'sd',
 'sdtautau': 'sd',
 'sdsd': 'sdsd',
 'ucuc': 'cucu',
 'suenue': 'usenu',
 'suenumu': 'usenu',
 'suenutau': 'usenu',
 'sumunue': 'usmunu',
 'sumunumu': 'usmunu',
 'sumunutau': 'usmunu',
 'sutaunue': 'ustaunu',
 'sutaunumu': 'ustaunu',
 'sutaunutau': 'ustaunu'}


class WilsonCoefficients(wilson.Wilson):
    """Class representing a point in the EFT parameter space and giving
    access to RG evolution.

    Note that all Wilson coefficient values refer to new physics contributions
    only, i.e. they vanish in the SM.

    Methods:

    - set_initial: set the initial values of Wilson coefficients at some scale
    - get_wc: get the values of the Wilson coefficients at some scale
    - set_initial_wcxf: set the initial values from a wcxf.WC instance
    - get_wc_wcxf: get the values of the Wilson coefficients at some scale
      as a wcxf.WC instance
    """
    def __init__(self):
        self.wc = None
        self._options = {}

    def set_initial(self, wc_dict, scale, eft='WET', basis='flavio'):
        """Set initial values of Wilson coefficients.

        Parameters:

        - wc_dict: dictionary where keys are Wilson coefficient name strings and
          values are Wilson coefficient NP contribution values
        - scale: $\overline{\text{MS}}$ renormalization scale
        """
        super().__init__(wcdict=wc_dict, scale=scale, eft=eft, basis=basis)

    def set_initial_wcxf(self, wc):
        """Set initial values of Wilson coefficients from a WCxf WC instance.

        If the instance is given in a basis other than the flavio basis,
        the translation is performed automatically, if implemented in the
        `wcxf` package."""
        super().__init__(wcdict=wc.dict, scale=wc.scale, eft=wc.eft, basis=wc.basis)

    @property
    def get_initial_wcxf(self):
        """Return a wcxf.WC instance in the flavio basis containing the initial
        values of the Wilson coefficients."""
        if self.wc is None:
            raise ValueError("Need to set initial values first.")
        return self.wc

    @classmethod
    def from_wilson(cls, w, par_dict):
        if w is None:
            return None
        if isinstance(w, cls):
            return w
        fwc = cls()
        fwc.set_initial_wcxf(w.wc)
        fwc._cache = w._cache
        fwc._options = w._options
        _ckm_options = {k: par_dict[k] for k in ['Vus', 'Vcb', 'Vub', 'delta']}
        if fwc.get_option('parameters') != _ckm_options:
            fwc.set_option('parameters', _ckm_options)
        return fwc

    def run_wcxf(*args, **kwargs):
        raise ValueError("The method run_wcxf has been removed. Please use the match_run method of wilson.Wilson instead.")

    def get_wcxf(self, sector, scale, par, eft='WET', basis='flavio', nf_out=None):
        """Get the values of the Wilson coefficients belonging to a specific
        sector (e.g. `bsmumu`) at a given scale.

        Returns a WCxf.WC instance.

        Parameters:

        - sector: string name of the sector as defined in the WCxf EFT instance
        - scale: $\overline{\text{MS}}$ renormalization scale
        - par: dictionary of parameters
        - eft: name of the EFT at the output scale
        - basis: name of the output basis
        """
        # nf_out is only present to preserve backwards compatibility
        if nf_out == 5:
            eft = 'WET'
        elif nf_out == 4:
            eft = 'WET-4'
        elif nf_out == 3:
            eft = 'WET-3'
        elif nf_out is not None:
            raise ValueError("Invalid value: nf_out=".format(nf_out))
        if sector == 'all':
            mr_sectors = 'all'
        else:
            # translate from legacy flavio to wcxf sector if necessary
            wcxf_sector = sectors_flavio2wcxf.get(sector, sector)
            mr_sectors = (wcxf_sector,)
        if not self.wc:
            return wcxf.WC(eft=eft, basis=basis, scale=scale, values={})
        return self.match_run(scale=scale, eft=eft, basis=basis, sectors=mr_sectors)

    def get_wc(self, sector, scale, par, eft='WET', basis='flavio', nf_out=None):
        """Get the values of the Wilson coefficients belonging to a specific
        sector (e.g. `bsmumu`) at a given scale.

        Returns a dictionary of WC values.

        Parameters:

        - sector: string name of the sector as defined in the WCxf EFT instance
        - scale: $\overline{\text{MS}}$ renormalization scale
        - par: dictionary of parameters
        - eft: name of the EFT at the output scale
        - basis: name of the output basis
        """
        wcxf_basis = wcxf.Basis[eft, basis]
        if sector == 'all':
            coeffs = wcxf_basis.all_wcs
        else:
            # translate from legacy flavio to wcxf sector if necessary
            wcxf_sector = sectors_flavio2wcxf.get(sector, sector)
            coeffs = wcxf_basis.sectors[wcxf_sector].keys()
        wc_sm = dict.fromkeys(coeffs, 0)
        if not self.wc or not any(self.wc.values.values()):
            return wc_sm
        wc_out = self.get_wcxf(sector, scale, par, eft, basis, nf_out)
        wc_out_dict = wc_sm  # initialize with zeros
        wc_out_dict.update(wc_out.dict)  # overwrite non-zero entries
        return wc_out_dict


# this global variable is simply an instance that is not meant to be modifed -
# i.e., a Standard Model Wilson coefficient instance.
_wc_sm = WilsonCoefficients()



r"""Functions for observables in $K\to\pi\pi$ decays, in particular
$\epsilon'/\epsilon$."""

import flavio
from flavio.classes import Prediction, Observable
from flavio.config import config
from flavio.physics import ckm
from flavio.physics.kdecays.wilsoncoefficients import wilsoncoefficients_sm_fourquark
from math import sqrt



def Kpipi_matrixelements_SM(par, scale):
    r"""Return the $K\to\pi\pi$ matrix elements of the SM operators in the
    traditional 10-operator basis.

    Returns a dictionary with keys 0 and 2 corresponding to the $\Delta I=1/2$
    and 3/2 matrix elements, respectively, and values being again dictionaries
    with the keys being the operator index and the value being the matrix
    element in units of GeV³, using the same normalization as in
    arXiv:1502.00263."""
    M0 = {}
    M2 = {}
    for i in [3, 4, 5, 6, 7, 8, 9]:
        M0[str(i)] = par['Kpipi M0 {}'.format(i)]
    for i in [7, 8, 9]:
        M2[str(i)] = par['Kpipi M2 {}'.format(i)]
    # Exact relations
    for i in [3, 4, 5, 6]:
        M2[str(i)] = 0
    M0['1'] = 1 / 3 * (M0['3'] + 2 * M0['9'])
    M0['2'] = 1 / 3 * (-2 * M0['3'] + 3 * M0['4'] + 2 * M0['9'])
    M0['10'] = -M0['3'] + M0['4'] + M0['9']
    M2['1'] = 2 / 3 * M2['9']
    M2['2'] = M2['1']
    M2['10'] = M2['9']
    return {0: M0, 2: M2}


def Kpipi_matrixelements_NP(par, scale):
    r"""Return the $K\to\pi\pi$ matrix elements of all $s\to d$ operators in the
    flavio basis.

    Returns a dictionary with keys 0 and 2 corresponding to the $\Delta I=1/2$
    and 3/2 matrix elements, respectively, and values being again dictionaries
    with the keys being the operator name and the value being the matrix
    element in units of GeV³, using the same normalization as in
    arXiv:1502.00263."""
    MSM = Kpipi_matrixelements_SM(par, scale)
    M = {0: {}, 2: {}}
    ms = flavio.physics.running.running.get_ms(par, scale, nf_out=3)
    # follows appendix A4 of Aebischer/Buras/Gerard arXiv:1807.01709
    for i in (0, 2):
        M[i]['CVLL_sduu']  =  MSM[i]['1'] / 4
        M[i]['CVLLt_sduu'] =  MSM[i]['2'] / 4
        M[i]['CVLR_sduu'] = (MSM[i]['5'] / 3 + 2 * MSM[i]['7'] / 3) / 4
        M[i]['CVLRt_sduu'] = (MSM[i]['6'] / 3 + 2 * MSM[i]['8'] / 3) / 4
        M[i]['CVLL_sddd'] = (2 * MSM[i]['3'] / 3 - 2 * MSM[i]['9'] / 3) / 4
        M[i]['CVLR_sddd'] = (2 * MSM[i]['5'] / 3 - 2 * MSM[i]['7'] / 3) / 4
        M[i]['CSRL_sddd'] = (MSM[i]['6'] / 3 - MSM[i]['8'] / 3) / 4
        M[i]['CSRR_sddd']  =  par['Kpipi M{} SLL2_d'.format(i)]
        M[i]['CTRR_sddd']  = -8 * par['Kpipi M{} SLL1_d'.format(i)] -4 * par['Kpipi M{} SLL2_d'.format(i)]
    for i in (2, ):
        # isospin relations valid for I=2 amplitude
        M[i]['CSRL_sduu'] = -M[i]['CSRL_sddd']
        M[i]['CSRR_sduu'] = -M[i]['CSRR_sddd']
        M[i]['CTRR_sduu'] = -M[i]['CTRR_sddd']
        M[i]['CSRLt_sduu'] = 1 / 2 * M[i]['CVLR_sddd']  # -1 from isospin, -1/2 from Fierz
        M[i]['CSRRt_sduu'] = -(-1 / 2 * M[i]['CSRR_sddd'] - 1 / 8 * M[i]['CTRR_sddd'])
        M[i]['CTRRt_sduu'] = -(-6 * M[i]['CSRR_sddd'] + 1 / 2 * M[i]['CTRR_sddd'])
        M[i]['C8_sd']      = 0
    for i in (0, ):
        M[i]['CSRL_sduu']  =  par['Kpipi M{} SLR2_u'.format(i)]
        M[i]['CSRLt_sduu'] =  par['Kpipi M{} SLR1_u'.format(i)]
        M[i]['CSRR_sduu']  =  par['Kpipi M{} SLL2_u'.format(i)]
        M[i]['CSRRt_sduu'] =  par['Kpipi M{} SLL1_u'.format(i)]
        M[i]['CTRR_sduu']  = -par['Kpipi M{} SLL4_u'.format(i)]
        M[i]['CTRRt_sduu'] = -par['Kpipi M{} SLL3_u'.format(i)]
        M[i]['C8_sd']      = -ms * par['Kpipi M0 g-'] / 2
    for i in (0, 2):
        M[i]['CVRR_sduu']  = -M[i]['CVLL_sduu']
        M[i]['CVRRt_sduu'] = -M[i]['CVLLt_sduu']
        M[i]['CVRR_sddd']  = -M[i]['CVLL_sddd']
        M[i]['CVRL_sduu']  = -M[i]['CVLR_sduu']
        M[i]['CVRLt_sduu'] = -M[i]['CVLRt_sduu']
        M[i]['CVRL_sddd']  = -M[i]['CVLR_sddd']
        M[i]['CSLR_sduu']  = -M[i]['CSRL_sduu']
        M[i]['CSLRt_sduu'] = -M[i]['CSRLt_sduu']
        M[i]['CSLR_sddd']  = -M[i]['CSRL_sddd']
        M[i]['CSLL_sduu']  = -M[i]['CSRR_sduu']
        M[i]['CSLLt_sduu'] = -M[i]['CSRRt_sduu']
        M[i]['CSLL_sddd']  = -M[i]['CSRR_sddd']
        M[i]['CTLL_sduu']  = -M[i]['CTRR_sduu']
        M[i]['CTLLt_sduu'] = -M[i]['CTRRt_sduu']
        M[i]['CTLL_sddd']  = -M[i]['CTRR_sddd']
        M[i]['C8p_sd']     = -M[i]['C8_sd']
    return M


def Kpipi_amplitudes_SM(par,
                       include_VmA=True, include_VpA=True,
                       scale_ImA0EW=False):
    r"""Compute the SM contribution to the two isospin amplitudes of
    the $K\to\pi\pi$ transition."""
    scale = config['renormalization scale']['kpipi']
    pref = par['GF'] / sqrt(2) * ckm.xi('u', 'ds')(par)  # GF/sqrt(2) Vus* Vud
    me = Kpipi_matrixelements_SM(par, scale)
    # Wilson coefficients
    wc = wilsoncoefficients_sm_fourquark(par, scale)
    tau = -ckm.xi('t', 'ds')(par) / ckm.xi('u', 'ds')(par)
    k = [1, 2]
    if include_VmA:
        k = k + [3, 4, 9, 10]
    if include_VpA:
        k = k + [5, 6, 7, 8]
    A = {0: 0, 2: 0}
    for i in [0, 2]:
        for j in k:
            m = me[i][str(j)]
            yj = wc.get('y{}'.format(j), 0)
            zj = wc.get('z{}'.format(j), 0)
            dA = pref * m * (zj + tau * yj)
            if scale_ImA0EW and i == 0 and j in [7, 8, 9, 10]:
                b = 1 / par['epsp a'] / (1 - par['Omegahat_eff'])
                dA = dA.real + 1j * b * dA.imag
            A[i] += dA
    return A


def Kpipi_amplitudes_NP(wc_obj, par):
    r"""Compute the new physics contribution to the two isospin amplitudes
    of the $K\to\pi\pi$ transition."""
    scale = config['renormalization scale']['kpipi']
    pref = 4 * par['GF'] / sqrt(2) * ckm.xi('t', 'ds')(par)  # 4GF/sqrt(2) Vts* Vtd
    me = Kpipi_matrixelements_NP(par, scale)
    wc = wc_obj.get_wc(sector='sd', scale=scale, par=par, eft='WET-3')
    A = {0: 0, 2: 0}
    for i in [0, 2]:
        for j, m in me[i].items():
            A[i] += -pref * m * complex(wc[j]).conjugate()  # conjugate!
    return A

def epsprime_SM(par):
    r"""Compute the SM contribution to $\epsilon'/\epsilon$, including
    isospin breaking corrections."""
    a = par['epsp a']
    A = Kpipi_amplitudes_SM(par)
    ImA0 = A[0].imag
    ImA2 = A[2].imag
    ReA0 = par['ReA0(K->pipi)']
    ReA2 = par['ReA2(K->pipi)']
     # eq. (19) of arXiv:1507.06345
    return (-par['omega+'] / (sqrt(2) * par['eps_K'])
            * (ImA0 / ReA0 * (1 - par['Omegahat_eff'])
               - 1 / a * ImA2 / ReA2).real)


def epsprime_NP(wc_obj, par):
    r"""Compute the NP contribution to $\epsilon'/\epsilon$."""
    # Neglecting isospin breaking corrections!
    A = Kpipi_amplitudes_NP(wc_obj, par)
    ImA0 = A[0].imag
    ImA2 = A[2].imag
    ReA0 = par['ReA0(K->pipi)']
    ReA2 = par['ReA2(K->pipi)']
    a = par['epsp a']  # eq. (16)
    # dividing by a to remove the isospin brk corr in omega+, cf. (16) in 1507.06345
    return (-par['omega+'] / a / (sqrt(2) * par['eps_K'])
            * (ImA0 / ReA0 - ImA2 / ReA2).real)

def epsprime(wc_obj, par):
    r"""Compute $\epsilon'/\epsilon$, parametrizing direct CPV in
    $K\to\pi\pi$."""
    return epsprime_SM(par) + epsprime_NP(wc_obj, par)


# Observable and Prediction instances
o = Observable('epsp/eps')
o.tex = r"$\epsilon^\prime/\epsilon$"
Prediction('epsp/eps', epsprime)
o.set_description(r"Direct CP violation parameter")
o.add_taxonomy(r'Process :: $s$ hadron decays :: Non-leptonic decays :: $K\to \pi\pi$')



"""This file contains the functions that allow to regenerate the data files
for the interpolation of the QCDF corrections. It is normally not needed."""

import flavio
import numpy as np
from flavio.physics.bdecays.bvll import qcdf
from flavio.config import config
from flavio.physics.bdecays.common import meson_quark
from flavio.physics.common import conjugate_par

# we will interpolate from 0 to 9 GeV^2 (above, QCDF is not applicable)
# shifted by 10^-6 to avoid the q2=0 singularity but be below the e+e- threshold
q2_arr = np.arange(0 + 1e-6, 9 + 1e-6, 0.1)

# NP contributions to QCDF corrections will be neglected throughout!
wcsm = flavio.WilsonCoefficients()

# central default values are assumed by all parameters

par = flavio.default_parameters.copy()
par_dict = par.get_central_all()
par_dict_cpconj = conjugate_par(par_dict)


array_dict = {}
i = 0
contributions_dict = {
'all': {'include_WA': True,  'include_O8': True,  'include_QSS': True},
'WA':  {'include_WA': True,  'include_O8': False, 'include_QSS': False},
'O8':  {'include_WA': False, 'include_O8': True,  'include_QSS': False},
'QSS': {'include_WA': False, 'include_O8': False, 'include_QSS': True },
}
for had in [('B0','K*0'), ('B+','K*+'), ('Bs','phi')]: #, ('B0','rho0'), ('B+','rho+') ]:
    for cp_conjugate in [False, True]: # compute it for the decay and its CP conjugate
        for contribution_name, contribution_dict in contributions_dict.items():
            l = 'e'
            process = had[0] + '->' + had[1] # e.g. B0->K*0
            print('Computing ' + process + ', contribution: ' + contribution_name + ' (' + str(i+1) + '/' + str(2*3*4) + ')')
            # compute corrections for each q2 value
            scale = config['renormalization scale']['bvll']
            label = meson_quark[had] + l + l # e.g. bsee
            wcsm_dict =  flavio.physics.bdecays.wilsoncoefficients.wctot_dict(wcsm, label, scale, par_dict)
            if cp_conjugate:
                hel_amps_list = [ qcdf.helicity_amps_qcdf(q2, wcsm_dict, par_dict_cpconj, B=had[0], V=had[1], **contribution_dict)
                              for q2 in q2_arr]
            else:
                hel_amps_list = [ qcdf.helicity_amps_qcdf(q2, wcsm_dict, par_dict, B=had[0], V=had[1],  **contribution_dict)
                              for q2 in q2_arr]
            # turn into numpy array
            hel_amps_arr = np.array([
                [hel_amps[('pl','V')],
                 hel_amps[('mi','V')],
                 hel_amps[('0','V')]]
                for hel_amps in hel_amps_list
            ])
            # divide out the q^2 poles to get a saner interpolant
            hel_amps_arr[:,0] = q2_arr * hel_amps_arr[:,0]
            hel_amps_arr[:,1] = q2_arr * hel_amps_arr[:,1]
            hel_amps_arr[:,2] = np.sqrt(q2_arr) * hel_amps_arr[:,2]
            if cp_conjugate:
                array_name = process + ' CP conjugate ' + contribution_name
            else:
                array_name = process + ' ' + contribution_name
            array_dict[array_name] = hel_amps_arr
            i += 1

np.savez('qcdf_interpolate', **array_dict)



r"""Functions for $\tau\to P\ell$."""

import flavio
from math import sqrt, pi
from flavio.physics.common import lambda_K
from flavio.physics.taudecays import common
from flavio.physics import ckm
from flavio.physics.running import running

# names of LFV sectors in WCxf
wcxf_sector_names = {('tau', 'mu'): 'mutau',
                     ('tau', 'e'): 'taue',
                     ('mu', 'e'): 'mue', }

def br_taupl(wc_obj, par, P, lep):
    r"""Branching ratio of $\tau^+\to p^0\ell^+$."""
    scale = flavio.config['renormalization scale']['taudecays']
    GF = par['GF']
    alphaem = running.get_alpha(par, 4.8)['alpha_e']
    mP = par['m_' + P]
    mtau = par['m_tau']
    mu = par['m_u']
    md = par['m_d']
    mlep = par['m_' + lep]
    fP = par['f_' + P]  

    if P == 'pi0':
        sec = wcxf_sector_names['tau', lep]
        wc = wc_obj.get_wc(sec, scale, par, nf_out=4)
        F = {}
        S = {}
        for q in 'ud':
            F['LL' + q] = wc['CVLL_tau{}{}'.format(lep, 2 * q)]  # FLL
            F['RR' + q] = wc['CVRR_tau{}{}'.format(lep, 2 * q)]  # FRR
            F['LR' + q] = wc['CVLR_tau{}{}'.format(lep, 2 * q)]  # FLR
            F['RL' + q] = wc['CVLR_{}tau{}'.format(2 * q, lep)]  # FLRqq
     
            S['RR' + q] = wc['CSRR_tau{}{}'.format(lep, 2 * q)]  # SRR
            S['RL' + q] = wc['CSRL_tau{}{}'.format(lep, 2 * q)]  # SLR
            S['LL' + q] = wc['CSRR_{}tau{}'.format(lep, 2 * q)].conjugate()  # SRRqq
            S['LR' + q] = wc['CSRL_{}tau{}'.format(lep, 2 * q)].conjugate()  # SLRqq

        vL = fP*((F['LRu'] - F['LLu'])/2  - (F['LRd'] - F['LLd'])/2 )/sqrt(2)
        vR = fP*((F['RRu'] - F['RLu'])/2  - (F['RRd'] - F['RLd'])/2 )/sqrt(2)
        sL = fP*mP**2/(mu+md)*((S['LRu']-S['LLu'])/2 - (S['LRd']-S['LLd'])/2)/sqrt(2)
        sR = fP*mP**2/(mu+md)*((S['RRu']-S['RLu'])/2 - (S['RRd']-S['RLd'])/2)/sqrt(2) 
        gL = sL+ (-vL*mlep + vR*mtau)
        gR = sR+ (-vR*mlep + vL*mtau)

    elif P == 'K0':
       xi_t = ckm.xi('t','ds')(par)
       qqll = 'sdtau' + lep
       wc = wc_obj.get_wc(qqll, scale, par, nf_out=4)
       mq1 = flavio.physics.running.running.get_md(par, scale)
       mq2 = flavio.physics.running.running.get_ms(par, scale)

       C9m =  -wc['C9_'+qqll]  + wc['C9p_'+qqll]
       C10m = -wc['C10_'+qqll] + wc['C10p_'+qqll]
       CSm =   wc['CS_'+qqll]  - wc['CSp_'+qqll]
       CPm =   wc['CP_'+qqll]  - wc['CPp_'+qqll]
       Kv = xi_t * 4*GF/sqrt(2) * alphaem/(4*pi)
       Ks = Kv * mq2
       gV = - fP* Kv * C9m/2
       gA = - fP* Kv * C10m/2
       gS = - fP* Ks * mP**2 * CSm/2 / (mq1+mq2)
       gP = - fP* Ks * mP**2 * CPm/2 / (mq1+mq2)
       gVS = gS + gV*(mtau -mlep)
       gAP = gP - gA*(mtau +mlep)
       gL =  gVS - gAP
       gR =  gVS + gAP
    return par['tau_tau'] * common.GammaFsf(mtau, mP, mlep, gL, gR)

# function returning function needed for prediction instance
def br_taupl_fct(P, lep):
    def f(wc_obj, par):
        return br_taupl(wc_obj, par, P, lep)
    return f

# Observable and Prediction instances

_had = {'pi0': r'\pi^0', 'K0' : r'K^0'}
_shortname = {'pi0': 'pi', 'K0': 'K'}
_lep = {'e': ' e', 'mu': r'\mu',}

for P in _had:
    for l in _lep:
        _obs_name = "BR(tau->" + _shortname[P] + l + r")"
        _obs = flavio.classes.Observable(_obs_name)
        _process_tex = r"\tau^+\to " + _had[P] + _lep[l] + r"^+"
        _process_taxonomy = r'Process :: $\tau$ lepton decays :: LFV decays :: $\tau\to P\ell$ :: $' + _process_tex + r"$"
        _obs.add_taxonomy(_process_taxonomy)
        _obs.set_description(r"Branching ratio of $" + _process_tex + r"$")
        _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
        flavio.classes.Prediction(_obs_name, br_taupl_fct(P, l))



r"""$W\to\ell\nu$ decays."""


import flavio
from flavio.physics.zdecays import smeftew
from math import sqrt, pi


def gWl_SM(f1, f2, par):
    if f1 == f2:
        return 1 / 2   # using neutrino flavour eigenstates
    else:
        return 0

def gWq_SM(f1, f2, par):
    i = smeftew._sectors['u'][f1]
    j = smeftew._sectors['d'][f2]
    V = flavio.physics.ckm.get_ckm(par)
    return V[i - 1, j - 1] / 2

def gWf_SM(f1, f2, par):
    if f1 in smeftew._sectors['l']:
        return gWl_SM(f1, f2, par)
    else:
        return gWq_SM(f1, f2, par)

def GammaWq_SM(f1, f2, par):
    return 4 * abs(gWq_SM(f1, f2, par))**2 * par['GammaW_had'] / 2

def GammaWl_SM(f1, f2, par):
    return 4 * abs(gWl_SM(f1, f2, par))**2 * par['GammaW_lep'] / 3

def GammaW_SM(f1, f2, par):
    if f1 in smeftew._sectors['l']:
        return GammaWl_SM(f1, f2, par)
    if f1 in smeftew._sectors['u']:
        return GammaWq_SM(f1, f2, par)


def GammaW_NP(par, Nc, gW_SM, d_gW):
    GF, mW = par['GF'], par['m_W']
    return (sqrt(2) * GF * mW**3 / (3 * pi) * Nc
            * (2*(gW_SM*d_gW).real + abs(d_gW)**2))


def GammaW(wc_obj, par, f1, f2):
    scale = flavio.config['renormalization scale']['wdecays']
    wc_dict = wc_obj.get_wcxf(sector='all', scale=scale, par=par,
                              eft='SMEFT', basis='Warsaw')
    Nc = smeftew._QN[f1]['Nc']
    gW_SM = gWf_SM(f1, f2, par)
    GSM = GammaW_SM(f1, f2, par)
    d_gW = smeftew.d_gW(f1, f2, par, wc_dict)
    GNP = GammaW_NP(par, Nc, gW_SM, d_gW)
    return GSM + GNP


def Gammalep(wc_obj, par):
    lep = ['e', 'mu', 'tau']
    return sum([GammaW(wc_obj, par, l1, l2) for l1 in lep for l2 in lep])

def Gammahad(wc_obj, par):
    return sum([GammaW(wc_obj, par, q1, q2) for q1 in 'uc' for q2 in 'dsb'])

def Gammatot(wc_obj, par):
    return Gammalep(wc_obj, par) + Gammahad(wc_obj, par)


def BRWlnu_fct(l):
    def fu(wc_obj, par):
        lep = ['e', 'mu', 'tau']
        # use calculated lifetime, which is more precise than measured one!
        return sum([GammaW(wc_obj, par, l, nu) for nu in lep]) / Gammatot(wc_obj, par)
    return fu


def RWlnu_fct(l1, l2):
    def fu(wc_obj, par):
        lep = ['e', 'mu', 'tau']
        # use calculated lifetime, which is more precise than measured one!
        num = sum([GammaW(wc_obj, par, l1, nu) for nu in lep])
        den = sum([GammaW(wc_obj, par, l2, nu) for nu in lep])
        return num / den
    return fu


def RWcX(wc_obj, par):
    q = ['d', 's', 'b']
    return sum([GammaW(wc_obj, par, 'c', d) for d in q]) / Gammahad(wc_obj, par)


_leptons = {'e': ' e', 'mu': r'\mu', 'tau': r'\tau'}
_uquarks = {'u': ' u', 'c': ' c', 'c': ' c'}
_dquarks = {'d': ' d', 's': ' s', 'b': ' b'}


for f, tex in _leptons.items():
    _process_tex = r"W^\pm\to {}^\pm\nu".format(tex)
    _obs_name = "BR(W->{}nu)".format(f)
    _obs = flavio.classes.Observable(_obs_name)
    _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
    _obs.set_description(r"Branching ratio of $" + _process_tex + r"$, summed over neutrino flavours")
    _obs.add_taxonomy(r'Process :: $W^\pm$ decays :: Leptonic decays :: $' + _process_tex + r"$")
    flavio.classes.Prediction(_obs_name, BRWlnu_fct(f))

for (f1, f2) in [('tau', 'e'), ('mu', 'e'), ('tau', 'mu')]:
    tex1 = _leptons[f1]
    tex2 = _leptons[f2]
    _process_tex = r"W^\pm\to \ell^\pm\nu".format(tex2)
    _process_tex1 = r"W^\pm\to {}^\pm\nu".format(tex1)
    _process_tex2 = r"W^\pm\to {}^\pm\nu".format(tex2)
    _obs_name = "R{}{}(W->lnu)".format(f1, f2)
    _obs = flavio.classes.Observable(_obs_name)
    _obs.tex = r"$\text{R}_{" + tex1 + " " + tex2 + "}(" + _process_tex + r")$"
    _obs.set_description(r"Ratio of branching ratio of $" + _process_tex1 + r"$ and  $" + _process_tex2 + r"$, individually summed over neutrino flavours")
    _obs.add_taxonomy(r'Process :: $W^\pm$ decays :: Leptonic decays :: $' + _process_tex1 + r"$")
    _obs.add_taxonomy(r'Process :: $W^\pm$ decays :: Leptonic decays :: $' + _process_tex2 + r"$")
    flavio.classes.Prediction(_obs_name, RWlnu_fct(f1, f2))


_process_tex = r"W^+\to cX"
_obs_name = "R(W->cX)"
_obs = flavio.classes.Observable(_obs_name)
_obs.tex = r"$\text{R}(" + _process_tex + r")$"
_obs.set_description(r"Ratio of partial width of $" + _process_tex + r"$, $X=\bar d, \bar s, \bar b$ over the hadronic $W$ width")
_obs.add_taxonomy(r'Process :: $W^\pm$ decays :: Hadronic decays :: $' + _process_tex + r"$")
flavio.classes.Prediction(_obs_name, RWcX)


_obs_name = "GammaW"
_obs = flavio.classes.Observable(_obs_name)
_obs.tex = r"$\Gamma_W$"
_obs.set_description(r"Total width of the $W^\pm$ boson")
for f, tex in _leptons.items():
    _process_tex = r"W^\pm\to {}^\pm\nu".format(tex)
    _obs.add_taxonomy(r'Process :: $W^\pm$ decays :: Leptonic decays :: $' + _process_tex + r"$")
flavio.classes.Prediction(_obs_name, Gammatot)



from setuptools import setup, find_packages

with open('flavio/_version.py', encoding='utf-8') as f:
    exec(f.read())

with open('README.md', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

setup(name='flavio',
      version=__version__,
      author='David M. Straub',
      author_email='david.straub@tum.de',
      url='https://flav-io.github.io',
      description='A Python package for flavour physics phenomenology in the Standard Model and beyond',
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      license='MIT',
      packages=find_packages(),
      package_data={
      'flavio':['data/*.yml',
                'data/test/*',
                'physics/data/arXiv-0810-4077v3/*',
                'physics/data/arXiv-1503-05534v1/*',
                'physics/data/arXiv-1503-05534v2/*',
                'physics/data/arXiv-1501-00367v2/*',
                'physics/data/arXiv-1602-01399v1/*',
                'physics/data/arXiv-1602-01399v1/*',
                'physics/data/arXiv-1811-00983v1/*',
                'physics/data/pdg/*',
                'physics/data/qcdf_interpolate/*',
                'physics/data/wcsm/*',
                ]
      },
      install_requires=['numpy', 'scipy', 'setuptools>=3.3', 'pyyaml',
                        'ckmutil', 'wilson>=1.7', ],
      extras_require={
            'testing': ['nose'],
            'plotting': ['matplotlib>=1.4'],
            'sampling': ['pypmc>=1.1', 'emcee', 'iminuit',],
            },
    )



import unittest
import flavio
from flavio.physics.taudecays import taulnunu
from wilson import Wilson
from math import sqrt


par = flavio.default_parameters.get_central_all()
wc_obj = flavio.WilsonCoefficients()


class TestTaulnunu(unittest.TestCase):
    def test_taulnunu_exp(self):
        # compare to the experimental values
        # cf. eq. (95) of 1705.00929
        self.assertAlmostEqual(flavio.sm_prediction('BR(tau->mununu)') / 0.1739,
                               1 / 1.0060, delta=0.0003)
        self.assertAlmostEqual(flavio.sm_prediction('BR(tau->enunu)') / 0.1782,
                               1 / 1.0022, delta=0.0003)

    def test_taulnunu_wrongflavor(self):
        self.assertEqual(taulnunu.BR_taulnunu(wc_obj, par, 'e', 'mu', 'e'), 0)

    def test_taulnunu_np(self):
        w = Wilson({'CVLL_numunuetaue': 2e-4, 'CVLL_numunueetau': 1e-4},
                    80, 'WET', 'flavio')
        wc_np = flavio.WilsonCoefficients.from_wilson(w, par)
        BR1 = taulnunu.BR_taulnunu(wc_np, par, 'e', 'e', 'mu')
        BR2 = taulnunu.BR_taulnunu(wc_np, par, 'e', 'mu', 'e')
        self.assertEqual(BR1 / BR2, 4)

    def test_taulnunu_np2(self):
        w = Wilson({'CVLL_numunuetaue': 2e-2, 'CVLL_numunuetaumu': 1e-2},
                    80, 'WET', 'flavio')
        BR1 = flavio.np_prediction('BR(tau->enunu)', w)
        BR2 = flavio.np_prediction('BR(tau->mununu)', w)
        self.assertAlmostEqual(BR1 / BR2, 4, delta=0.2)

    def test_GFeff(self):
        CLSM = -4 *  par['GF'] / sqrt(2)
        w = Wilson({'CVLL_numunueemu': CLSM},
                    80, 'WET', 'flavio')
        wc_np = flavio.WilsonCoefficients.from_wilson(w, par)
        GFeff = taulnunu.GFeff(wc_np, par)
        self.assertEqual(GFeff / par['GF'], 0.5)
        w = Wilson({'CVLL_numunueemu': -0.5 * CLSM},
                    80, 'WET', 'flavio')
        wc_np = flavio.WilsonCoefficients.from_wilson(w, par)
        GFeff = taulnunu.GFeff(wc_np, par)
        self.assertEqual(GFeff / par['GF'], 2)

    def test_taulnunu_np3(self):
        CLSM = -4 *  par['GF'] / sqrt(2)
        w = Wilson({'CVLL_numunueemu': CLSM},
                    80, 'WET', 'flavio')
        BRSM = flavio.sm_prediction('BR(tau->enunu)')
        BRNP = flavio.np_prediction('BR(tau->enunu)', w)
        self.assertEqual(BRNP / BRSM, 0.25)
        w = Wilson({'CVLL_numunueemu': -0.5 * CLSM},
                    80, 'WET', 'flavio')
        BRSM = flavio.sm_prediction('BR(tau->enunu)')
        BRNP = flavio.np_prediction('BR(tau->enunu)', w)
        self.assertEqual(BRNP / BRSM, 4)

    def test_taulnunu_np4(self):
        CLSM = -4 *  par['GF'] / sqrt(2)
        w = Wilson({'CVLL_numunueemu': -0.5 * CLSM,
                    'CVLL_nutaunueetau': CLSM},
                    80, 'WET', 'flavio')
        BR1 = flavio.sm_prediction('BR(tau->enunu)')
        BR2 = flavio.np_prediction('BR(tau->enunu)', w)
        self.assertEqual(BR2 / BR1, 9)



import unittest
import flavio
from .smeftew import *


class ZeroDict(dict):
    """Dictionary that always returns 0"""

    def __init__(self, arg):
        super().__init__()

    def __getitem__(self, arg):
        return 0


class TestSMEFTew(unittest.TestCase):
    def test_shifts_sm(self):
        C = ZeroDict({})
        par = flavio.default_parameters.get_central_all()
        self.assertEqual(d_GF(par, C), 0)
        self.assertEqual(d_mZ2(par, C), 0)
        self.assertEqual(d_gZb(par, C), 0)



from flavio.physics.bdecays.formfactors.lambdab_12 import sse
from flavio.classes import AuxiliaryQuantity, Implementation
from flavio.config import config

processes = ['Lambdab->Lambda',]

def ff_function(function, process, **kwargs):
    return lambda wc_obj, par_dict, q2: function(process, q2, par_dict, **kwargs)

for p in processes:
    quantity = p + ' form factor'
    a = AuxiliaryQuantity(name=quantity, arguments=['q2'])
    a.set_description('Hadronic form factor for the ' + p + ' transition')

    iname = p + ' SSE2'
    i = Implementation(name=iname, quantity=quantity,
                   function=ff_function(sse.ff, p, n=2))
    i.set_description("2-parameter simplified series expansion")

    iname = p + ' SSE3'
    i = Implementation(name=iname, quantity=quantity,
                   function=ff_function(sse.ff, p, n=2))
    i.set_description("3-parameter simplified series expansion")



r"""Functions for observables in exclusive semi-leptonic $B\to V\ell^+\ell^-$ FCNC decays."""


from . import amplitudes
from . import observables
from . import observables_bs
from . import qcdf
from . import qcdf_interpolate
from . import lfv
from . import nonfactorizable
from . import subleading



import numpy as np
import json
import pkgutil
from flavio.classes import Parameter
from flavio.statistics.probability import MultivariateNormalDistribution


FFs = ["A0", "A1", "A12", "V", "T1", "T2", "T23"]
ai = ["a0", "a1", "a2"]
ff_a = [(ff, a) for ff in FFs for a in ai]
a_ff_string = [a + '_' + ff for ff in FFs for a in ai]


tex_a = {'a0': 'a_0', 'a1': 'a_1', 'a2': 'a_2', }
tex_ff = {'A0': 'A_0', 'A1': 'A_1', 'A12': r'A_{12}', 'V': 'V', 'T1': 'T_1', 'T2': 'T_2', 'T23': r'T_{23}', }


def get_ffpar(filename):
    f = pkgutil.get_data('flavio.physics', filename)
    data = json.loads(f.decode('utf-8'))
    central = np.array([data['central'][ff].get(a, np.nan) for ff, a in ff_a])
    unc = np.array([data['uncertainty'][ff].get(a, np.nan) for ff, a in ff_a])
    corr = np.array([[data['correlation'][ff1 + ff2].get(a1 + a2, np.nan) for ff1, a1 in ff_a] for ff2, a2 in ff_a])
    # delete the parameters a0_A12 and a0_T1, which are instead fixed
    # using the exact kinematical relations, cf. eq. (16) of arXiv:1503.05534
    pos_a0_A12 = ff_a.index(('A12', 'a0'))
    pos_a0_T2 = ff_a.index(('T2', 'a0'))
    central = np.delete(central, [pos_a0_A12, pos_a0_T2])
    unc = np.delete(unc, [pos_a0_A12, pos_a0_T2])
    corr = np.delete(corr, [pos_a0_A12, pos_a0_T2], axis=0)
    corr = np.delete(corr, [pos_a0_A12, pos_a0_T2], axis=1)
    return [central, unc, corr]


def load_parameters(filename, process, constraints):
    implementation_name = process + ' BSZ'
    parameter_names = [implementation_name + ' ' + coeff_name for coeff_name in a_ff_string]
    # a0_A0 and a0_T2 are not treated as independent parameters!
    parameter_names.remove(implementation_name + ' a0_A12')
    parameter_names.remove(implementation_name + ' a0_T2')
    for parameter_name in parameter_names:
        try:  # check if parameter object already exists
            p = Parameter[parameter_name]
        except KeyError:  # if not, create a new one
            p = Parameter(parameter_name)
            # get LaTeX representation of coefficient and form factor names
            _tex_a = tex_a[parameter_name.split(' ')[-1].split('_')[0]]
            _tex_ff = tex_ff[parameter_name.split(' ')[-1].split('_')[-1]]
            p.tex = r'$' + _tex_a + r'^{' + _tex_ff + r'}$'
            p.description = r'BSZ form factor parametrization coefficient $' + _tex_a + r'$ of $' + _tex_ff + r'$'
        else:  # if parameter exists, remove existing constraints
            constraints.remove_constraint(parameter_name)
    [central, unc, corr] = get_ffpar(filename)
    constraints.add_constraint(parameter_names,
        MultivariateNormalDistribution(central_value=central, covariance=np.outer(unc, unc)*corr))


# Resonance masses used in arXiv:1503.05534
resonance_masses_bsz = {
    'B->K*': {
        'm0': 5.367,
        'm1-': 5.415,
        'm1+': 5.830,
    },
    'B->rho': {
        'm0': 5.279,
        'm1-': 5.324,
        'm1+': 5.716,
    },
    'B->omega': {
        'm0': 5.279,
        'm1-': 5.324,
        'm1+': 5.716,
    },
    'Bs->phi': {
        'm0': 5.367,
        'm1-': 5.415,
        'm1+': 5.830,
    },
    'Bs->K*': {
        'm0': 5.279,
        'm1-': 5.324,
        'm1+': 5.716,
    },
}

# Resonance masses used in arXiv:1811.00983
resonance_masses_gkvd = {
    'B->K*': {
        'm0': 5.336,
        'm1-': 5.412,
        'm1+': 5.829,
    },
    'B->rho': {
        'm0': 5.279,
        'm1-': 5.325,
        'm1+': 5.724,
    },
    'B->D*': {
        'm0': 6.275,
        'm1-': 6.330,
        'm1+': 6.767,
    },
}


def transition_filename(tr):
    """Get the part of the filename specifying the transition (e.g. BKstar)
    from a transition string (e.g. B->K*)."""
    return tr.replace('->', '').replace('*', 'star')


def bsz_load(version, fit, transitions, constraints):
    """Load the form factor parameters given in arXiv:1503.05534"""
    for tr in transitions:
        for m, v in resonance_masses_bsz[tr].items():
            constraints.set_constraint('{} BCL {}'.format(tr, m), v)
        filename = 'data/arXiv-1503-05534{}/{}_{}.json'.format(version, transition_filename(tr), fit)
        load_parameters(filename, tr, constraints)


def bsz_load_v1_lcsr(constraints):
    bsz_load('v1', 'LCSR', ('B->K*', 'B->omega', 'B->rho', 'Bs->phi', 'Bs->K*'), constraints)

def bsz_load_v1_combined(constraints):
    bsz_load('v1', 'LCSR-Lattice', ('B->K*', 'Bs->phi', 'Bs->K*'), constraints)

def bsz_load_v2_lcsr(constraints):
    bsz_load('v2', 'LCSR', ('B->K*', 'B->omega', 'B->rho', 'Bs->phi', 'Bs->K*'), constraints)

def bsz_load_v2_combined(constraints):
    bsz_load('v2', 'LCSR-Lattice', ('B->K*', 'Bs->phi', 'Bs->K*'), constraints)


def gkvd_load(version, fit, transitions, constraints):
    """Load the form factor parameters given in arXiv:1811.00983"""
    for tr in transitions:
        for m, v in resonance_masses_gkvd[tr].items():
            constraints.set_constraint('{} BCL {}'.format(tr, m), v)
        filename = 'data/arXiv-1811-00983{}/{}_{}.json'.format(version, transition_filename(tr), fit)
        load_parameters(filename, tr, constraints)



from flavio.physics.bdecays.formfactors.b_v import bsz, sse, cln, clnexp
from flavio.classes import AuxiliaryQuantity, Implementation
from flavio.config import config

processes_H2L = ['B->K*', 'B->rho', 'B->omega', 'Bs->phi', 'Bs->K*']  # heavy to light
processes_H2H = ['B->D*', ]  # heavy to heavy


def ff_function(function, process, **kwargs):
    return lambda wc_obj, par_dict, q2: function(process, q2, par_dict, **kwargs)


for p in processes_H2L + processes_H2H:

    quantity = p + ' form factor'
    a = AuxiliaryQuantity(name=quantity, arguments=['q2'])
    a.set_description('Hadronic form factor for the ' + p + ' transition')

    iname = p + ' BSZ2'
    i = Implementation(name=iname, quantity=quantity,
                   function=ff_function(bsz.ff, p, n=2))
    i.set_description("2-parameter BSZ parametrization (see arXiv:1503.05534)")

    iname = p + ' BSZ3'
    i = Implementation(name=iname, quantity=quantity,
                   function=ff_function(bsz.ff, p, n=3))
    i.set_description("3-parameter BSZ parametrization (see arXiv:1503.05534)")

    iname = p + ' SSE'
    i = Implementation(name=iname, quantity=quantity,
                   function=ff_function(sse.ff, p, n=2))
    i.set_description("2-parameter simplified series expansion")

for p in processes_H2H:

    iname = p + ' CLN'
    i = Implementation(name=iname, quantity=quantity,
                   function=ff_function(cln.ff, p, scale=config['renormalization scale']['bvll']))
    i.set_description("CLN parametrization")

    iname = p + ' CLNexp-IW'
    i = Implementation(name=iname, quantity=quantity,
                   function=ff_function(clnexp.ff, p, scale=config['renormalization scale']['bvll']))
    i.set_description("CLN-like parametrization as used by B factories"
                      " and using improved Isgur-Wise relations"
                      " for the tensor form factors")



import unittest
import flavio
import numpy as np
from cmath import exp, pi
from wilson import Wilson

from flavio.physics.taudecays.test_taulgamma import compare_BR


class TestMuEGamma(unittest.TestCase):
    def test_muegamma(self):
        self.assertEqual(flavio.sm_prediction('BR(mu->egamma)'), 0)
    def test_tauegamma_implementation(self):
        input_dict_list=[{
        'Cgamma_mue':np.random.random()*1e-8*exp(1j*2*pi*np.random.random()),
        'Cgamma_emu':np.random.random()*1e-8*exp(1j*2*pi*np.random.random()),
        } for i in range(10)]
        BRs = np.array([
            flavio.np_prediction(
                'BR(mu->egamma)',
                Wilson(input_dict, 100,  'WET', 'flavio')
            )
            for input_dict in input_dict_list
        ])
        compare_BRs = np.array([
            compare_BR(
                Wilson(input_dict, 100,  'WET', 'flavio'),
                'mu', 'e',
            )
            for input_dict in input_dict_list
        ])
        self.assertAlmostEqual(np.max(np.abs(1-BRs/compare_BRs)), 0, delta=0.005)
